/**
	$Id: RemoteSlotManager.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent.slotmanager;


import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TimerTask;

import net.sf.basedb.clients.jobagent.Agent;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.Job;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.jobagent.JobAgentConnection;

/**
	Slot manager implementation that requests slots from another
	job agent. The other job agent should be using the {@link MasterSlotManager}.
	Communication between the job agents is made on the remote control
	port on the master job agent. Make sure that the master slot manager
	allows connections from this host by including it's ip address in the 
	'agent.remotecontrol' setting.
	
	<p>
	Configuration options for this slot manager:

	<ul>
	<li>agent.slotmanager.remote.server: The host name/ip of the master job agent
	<li>agent.slotmanager.remote.port: The port number of the remote control port on
		the master job agent.
	</ul>

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (on, 09 apr 2014) $
*/
public class RemoteSlotManager
	implements SlotManager
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(RemoteSlotManager.class);

	private Agent agent;
	private String remoteHost;
	private int remotePort;
	private int connectTimeout;
	private Map<String, Slot> remoteSlots;
	private TimerTask pingerTask;
	
	public RemoteSlotManager()
	{}
	
	/*
		From the SlotManager interface
		------------------------------
	*/
	@Override
	public void init(Agent agent)
	{
		log.info("Initalizing remote slot manager");
		this.agent = agent;
		this.remoteHost = agent.getProperty("agent.slotmanager.remote.server");
		this.remotePort = Values.getInt(agent.getProperty("agent.slotmanager.remote.port"));
		this.connectTimeout = Config.getInt("agent.connection.timeout", 1000);
		this.remoteSlots = new HashMap<String, Slot>();
		this.pingerTask = Application.getScheduler().schedule(new RemoteSlotPinger(this), 20000, 20000, false);
	}

	@Override
	public void close()
	{
		log.info("Closing remote slot manager");
		pingerTask.cancel();
		pingerTask = null;
		remoteSlots.clear();
		remoteSlots = null;
		agent = null;
	}

	@Override
	public synchronized Slot getSlot(Job job)
	{
		JobAgentConnection conn = new JobAgentConnection(remoteHost, remotePort, connectTimeout);
		Slot slot = null;
		try
		{
			String cmd = "slotmanager://" + remoteHost + ":" + remotePort + "/get-slot?" + job.getEstimatedExecutionTime().name();
			String answer = conn.send(cmd);
			Map<String, String> tmp = JobAgentConnection.parseSimpleAnswer(answer);
			String id = tmp.get("Id");
			Job.ExecutionTime assigned = Job.ExecutionTime.valueOf(tmp.get("Slot"));
			slot = new Slot(id, assigned);
			remoteSlots.put(id, slot);
		}
		catch (IOException ex)
		{
			log.error("Could not get slot from remote host: " + remoteHost, ex);
		}
		finally
		{
			conn.close();
		}
		return slot;
	}

	@Override
	public synchronized void releaseSlot(Slot slot)
	{
		JobAgentConnection conn = new JobAgentConnection(remoteHost, remotePort, connectTimeout);
		try
		{
			String cmd = "slotmanager://" + slot.getId() + "@" + remoteHost + ":" + remotePort + "/release-slot";
			conn.send(cmd);
			remoteSlots.remove(slot.getId());
		}
		catch (IOException ex)
		{
			log.error("Could not remove slot from remote host: " + slot, ex);
		}
		finally
		{
			conn.close();
		}
	}
	// -----------------------------------------------
	
	/**
		Ping the the remote host with all active slots.
	*/
	void pingRemoteSlots()
	{
		Set<String> slots;
		synchronized (this)
		{
			slots = new HashSet<String>(remoteSlots.keySet());
		}
		log.debug("Number of remote slots: " + slots.size());
		if (slots.size() == 0) return;
		JobAgentConnection conn = new JobAgentConnection(remoteHost, remotePort, connectTimeout);
		try
		{
			for (String slotId : slots)
			{
				log.debug("Pinging remote slot: " + slotId);
				try
				{
					String cmd = "slotmanager://" + slotId + "@" + remoteHost + ":" + remotePort + "/ping-slot";
					conn.send(cmd);
				}
				catch (IOException ex)
				{
					log.error("Could not ping slot: " + slotId, ex);
				}
			}
		}
		finally
		{
			conn.close();
		}
	}
	
	/**
		Task that is scheduled at regular intervals to send ping commands
		to the remote host.
	*/
	static class RemoteSlotPinger
		extends TimerTask
	{
	
		private final RemoteSlotManager master;
		
		RemoteSlotPinger(RemoteSlotManager master)
		{
			this.master = master;
		}
	
		/*
			From the TimerTask interface
			---------------------------------
		*/
		@Override
		public void run()
		{
			log.debug("Pinging remote slots");
			master.pingRemoteSlots();
		}
		// ----------------------------------
	}

}
