/**
	$Id: MenuItemAction.java 6652 2014-12-09 11:48:21Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.menu;

import net.sf.basedb.clients.web.extensions.UnsafeScriptable;
import net.sf.basedb.util.extensions.Action;

/**
	An action for extensions to the menu system. The extension point
	will add menu items to the <code>Extensions</code> menu. 
	<p>
	There are three types of menu items:
	<ul>
	<li>A regular menu item, which is usually associated with some kind
		of action when the user clicks on it.
	<li>A separator item. It's just for visual appearance. It is 
		never associated with any action.
	<li>A submenu item, which opens up a submenu when the user
		selects it. A submenu must specify a {@link #getSubmenuId()}.
		Regular menu items can be put into the submenu item by specifying
		the same id for {@link #getMenuId()}.
	</ul>

	@author nicklas
	@version 2.7
	@base.modified $Date: 2014-12-09 12:48:21 +0100 (ti, 09 dec 2014) $
*/
@UnsafeScriptable
public interface MenuItemAction
	extends Action
{

	/**
		ID attribute of the main menu tag. You need to set
		this value to reference the menu item in a javascript.
		@return The ID, or null
		@since 3.3
	*/
	public String getId();
	
	/**
		Class attribute to add to the main menu tag. 
		@since 3.3
	*/
	public String getClazz();

	/**
		The type of menu item.
	*/
	public MenuType getType();
	
	/**
		The ID of the menu where this menu item belongs, or null if
		it belongs to the top-level menu.
	*/
	public String getMenuId();
	
	/**
		The ID of a new submenu. This value is required when the menu item is
		a {@link MenuType#SUBMENU}, and is ignored otherwise.
	*/
	public String getSubmenuId();
	
	/**
		Get the title of the menu. 
	*/
	public String getTitle();
	
	/**
		A tooltip that is shown as a yellow popup when the 
		user points the mouse at the menu item.
		@return The tooltip or null to not display any tooltip
	*/
	public String getTooltip();
	
	/**
		Get an absolute reference to an image that will be displayed in the menu.
		It is recommended that the image is 9 pixels wide and 12 pixels high to
		line up with the icons used by the BASE core menus.	
		
		@return A reference to an image, or null if no image should
			be used
	*/
	public String getIcon();
	
	/**
		Extra CSS style attributes that will be added to the
		<code>style</code> attribute of the menu item div.
		@return Style attributes, or null
	*/
	public String getStyle();
	
	/**
		If the menu item should be enabled or disabled.
	*/
	public boolean isEnabled();
	
	/**
		If the menu item should be visible or not.
	*/
	public boolean isVisible();
	
	public static enum MenuType
	{
		/**
			This is a real menu item.
		*/
		MENUITEM,
		
		/**
			A menu item that opens a submenu. Menu items of this type must
			provide a unique value for {@link MenuItemAction#getSubmenuId()}.
			A regular menu item can be placed inside a submenu by specifying the
			same value for {@link MenuItemAction#getMenuId()}.
		*/
		SUBMENU,
		
		/**
			This is a separator line. All properties except 
			{@link MenuItemAction#getId()},
			{@link MenuItemAction#getClazz()},
			{@link MenuItemAction#getStyle()} and
			{@link MenuItemAction#isVisible()} 
			are ignored.
		*/
		SEPARATOR;
	}
	
}
