/**
	$Id: BioPlateEventParticipant.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.BioPlateData;
import net.sf.basedb.core.data.BioPlateEventParticipantData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Holds information about a bioplate that is participating in a bioplate event.
	The bioplate may have a specific 'role' and an 'index' (if the ordering among
	multiple plates is important). The BASE core itself doesn't requires or makes
	any use of those attributes. It is also possible to link individual biomaterial
	events to a participant.
	<p>
	
	A bioplate participant object is formally a child item to the bioplate event,
	but users that has read permission to the bioplate also get read permission
	to the participant object.

	@author Nicklas
	@since 2.17
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (må, 22 maj 2017) $
*/
public class BioPlateEventParticipant
	extends ChildItem
{

	
	/**
		The type of item represented by this class.
		@see Item#BIOPLATEEVENTPARTICIPANT
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOPLATEEVENTPARTICIPANT;
	
	/**
		The maximum length of the role string that can be stored in the database.
		@see #setRole(String)
	*/
	public static final int MAX_ROLE_LENGTH = BioPlateEventParticipantData.MAX_ROLE_LENGTH;

	/**
		Creates a new participant for the given event and bioplate.
		@see BioPlateEvent#addParticipant(BioPlate, String, int)
	*/
	static BioPlateEventParticipant getNew(DbControl dc, BioPlateEvent event, BioPlate plate)
	{
		BioPlateEventParticipant participant = dc.newItem(BioPlateEventParticipant.class);
		participant.getData().setBioPlate(plate.getData());
		participant.getData().setEvent(event.getData());
		event.getData().getParticipants().add(participant.getData());
		participant.setIndex(event.getData().getParticipants().size());
		return participant;
	}
	
	/**
		Get a <code>BioPlateEventParticipant</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioPlateEventParticipant</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static BioPlateEventParticipant getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioPlateEventParticipant participant = dc.loadItem(BioPlateEventParticipant.class, id);
		if (participant == null) 
		{
			throw new ItemNotFoundException("BioPlateEventParticipant[id="+id+"]");
		}
		return participant;
	}
	
	/**
		Get a query configured to return all participants for a bioplate event.
		
		@param event The event to get participants for (null is not allowed)
		@return An {@link ItemQuery} object
	 	@throws InvalidUseOfNullException If the event parameter is null
	*/
	public static ItemQuery<BioPlateEventParticipant> getQuery(BioPlateEvent event)
		throws InvalidDataException
	{
		if (event == null) throw new InvalidUseOfNullException("event");
		ItemQuery<BioPlateEventParticipant> query = new ItemQuery<BioPlateEventParticipant>(BioPlateEventParticipant.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("event"),
				Hql.entity(event)
			)
		);
		return query;
	}

	/**
		Get a query configured to return all event participants for a bioplate.
		
		@param bioPlate The bioplate to get participants for (null is not allowed)
		@return An {@link ItemQuery} object
	 	@throws InvalidUseOfNullException If the bioplate parameter is null
	*/
	public static ItemQuery<BioPlateEventParticipant> getQuery(BioPlate bioPlate)
		throws InvalidDataException
	{
		if (bioPlate == null) throw new InvalidUseOfNullException("bioPlate");
		ItemQuery<BioPlateEventParticipant> query = new ItemQuery<BioPlateEventParticipant>(BioPlateEventParticipant.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("bioPlate"),
				Hql.entity(bioPlate)
			)
		);
		return query;
	}

	
	BioPlateEventParticipant(BioPlateEventParticipantData data)
	{
		super(data);
	}
	
	@Override
	BioPlateEventParticipantData getData()
	{
		return (BioPlateEventParticipantData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getEvent();
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	@Override
	void initPermissions(int granted, int denied) 
		throws BaseException
	{
		BioPlateData plate = getData().getBioPlate();
		int platePermission = Permission.READ.grantValue();
		if (plate.getId() != 0)
		{
			platePermission = getSessionControl().getAllPermissions(plate);
		}
		if (Permission.hasPermission(platePermission, Permission.READ))
		{
			granted |= Permission.grant(Permission.READ);
		}
		super.initPermissions(granted, denied);
	}

	@Override
	void onBeforeCommit(Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Action.DELETE)
		{
			// Unlink all biomaterial events from this participant
			org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedQuery(
				getDbControl().getHibernateSession(), "CLEAR_PARTICIPANT_ON_BIOMATERIALEVENTS_FOR_PARTICIPANT");
			/*
				UPDATE BioMaterialEventData bme
				SET bme.bioPlateEventParticipant = null
				WHERE bme.bioPlateEventParticipant = :participant
			*/
			query.setParameter("participant", getId(), TypeWrapper.H_INTEGER);
			query.executeUpdate();
		}
	}

	/**
		Get the {@link BioPlate}.
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the bioplate
		@throws BaseException If there is another error
	*/
	public BioPlate getBioPlate()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioPlate.class, getData().getBioPlate());
	}

	/**
		Get the {@link BioPlateEvent}.
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the event
		@throws BaseException If there is another error
	*/
	public BioPlateEvent getEvent()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioPlateEvent.class, getData().getEvent());
	}

	/**
		Get the role that the plate has in this event. The role is a string defined by 
		the client application.
	*/
	public String getRole()
	{
		return getData().getRole();
	}
	
	/**
		Set the role that the plate has in this event. The role is a string defined by 
		the client application.
	*/
	public void setRole(String role)
	{
		checkPermission(Permission.WRITE);
		getData().setRole(StringUtil.setNullableString(role, "role", MAX_ROLE_LENGTH));
	}
	
	/**
		Get the index of the plate in this event. This has no meaning unless the
		event require that the plates are ordered.
	*/
	public int getIndex()
	{
		return getData().getIndex();
	}
	
	/**
		Set the index of the plate in this event. This has no meaning unless the
		event require that the plates are ordered.
	*/
	public void setIndex(int index)
	{
		checkPermission(Permission.WRITE);
		getData().setIndex(index);
	}
}
