/*
	$Id: ExtraValue.java 6881 2015-04-21 07:59:42Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ExtraValueData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.HashMap;

/**
	Represents extra values attached to a <code>BioAssaySet</code>. 
	The type of extra values are defined by the {@link ExtraValueType}.
	The insertion of extra values is done using batchers, created by calling either {@link 
	BioAssaySet#getSpotExtraValueBatcher(Class, ExtraValueType, Job)}
	or {@link BioAssaySet#getPositionExtraValueBatcher(Class, ExtraValueType, Job)}.
	Only one set of extra values of the same <code>ExtraValueType</code> are
	allowed in a bioassayset.
	
	@author Nicklas
	@version 2.0
	@see SpotExtraValueBatcher
	@see PositionExtraValueBatcher
	@base.modified $Date: 2015-04-21 09:59:42 +0200 (ti, 21 apr 2015) $
*/
public class ExtraValue
	extends ChildItem
	implements FileStoreEnabled
{
	/**
		The type of item represented by this class.
		@see Item#EXTRAVALUE
		@see #getType()
	*/
	public static final Item TYPE = Item.EXTRAVALUE;


	/**
		Create a new <code>ExtraValue</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param bioAssaySet The bioassayset the extra value belongs to
		@param extraValueType The type of the extra value
		@param dcev The data cube extra value used to store the extra values
		@param job The job that calculated the extra values or null
		@return The new <code>ExtraValue</code> item
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for the experiment
		@throws InvalidDataException If the bioassayset or extra value type is null
		@throws BaseException If there is an error
	*/
	static ExtraValue getNew(DbControl dc, BioAssaySet bioAssaySet, ExtraValueType extraValueType, 
		DataCubeExtraValue dcev, Job job)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		ExtraValue ev = dc.newItem(ExtraValue.class);
		ev.setBioAssaySet(bioAssaySet);
		ev.setExtraValueType(extraValueType);
		ev.setDataCubeExtraValue(dcev);
		ev.setJob(job);
		return ev;
	}

	/**
		Get an <code>ExtraValue</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>ExtraValue</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static ExtraValue getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ExtraValue ev = dc.loadItem(ExtraValue.class, id);
		if (ev == null) throw new ItemNotFoundException("ExtraValue[id=" + id + "]");
		return ev;
	}
	
	/**
		Get a query configured to retrieve extra values in a given bioassayset.
		@param bioAssaySet The bioassayset to retreive bioassays for, null
			is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If the argument is null. 
		@see #getValues()
	*/
	public static ItemQuery<ExtraValue> getQuery(BioAssaySet bioAssaySet)
		throws InvalidDataException
	{
		if (bioAssaySet == null) throw new InvalidUseOfNullException("bioAssaySet");
		ItemQuery<ExtraValue> query = new ItemQuery<ExtraValue>(ExtraValue.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("bioAssaySet"),
				Hql.entity(bioAssaySet)
			)
		);
		return query;
	}

	/**
	 	Creates a new extra value item from the given data.
		@param data the data
	*/
	ExtraValue(ExtraValueData data)
	{
		super(data);
	}

	@Override
	ExtraValueData getData()
	{
		return (ExtraValueData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the FileStoreEnabled interface
		-------------------------------------------
	*/
	/**
		@since 2.8
	*/
	@Override
	public FileSet getFileSet()
	{
		DbControl dc = getDbControl();
		FileSet fs = dc.getItem(FileSet.class, getData().getFileSet(), this);
		if (fs == null)
		{
			fs = FileSet.getNew(dc, this);
			getData().setFileSet(fs.getData());
			dc.saveItemIf(this, fs, false);
		}
		return fs;
	}
	/**
		@since 2.8
	*/
	@Override
	public boolean hasFileSet()
	{
		return getData().getFileSet() != null;
	}
	/**
		@return Always null
		@since 2.8
	*/
	@Override
	public Platform getPlatform()
	{
		return null;
	}
	/**
		@return Always null
		@since 2.8
	*/
	@Override
	public PlatformVariant getVariant()
	{
		return null;
	}
	/**
		Get the file set from the parent bioassay set if it exists.
		@since 2.8
	*/
	@Override
	public Collection<FileSet> getParentFileSets()
	{
		BioAssaySet source = getBioAssaySet();
		Collection<FileSet> parents = null;
		if (source != null && source.hasFileSet())
		{
			parents = Collections.singleton(source.getFileSet());
		}
		return parents;
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getBioAssaySet().getExperiment();
	}
	/**
		USE permission is enough to be able to manage extra values.
	*/
	@Override
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	// -------------------------------------------

	/**
		Get the experiment this extra value belongs to.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getBioAssaySet().getExperiment());
	}
	
	/**
		Get the bioassayset this extra value belongs to.
		@return A <code>BioAssaySet</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the bioassayset
		@throws BaseException If there is another error
	*/
	public BioAssaySet getBioAssaySet()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioAssaySet.class, getData().getBioAssaySet());
	}
	
	/**
		Set the bioassayset. This cannot be changed after the extra value
		has been created.
	*/
	private void setBioAssaySet(BioAssaySet bioAssaySet)
		throws PermissionDeniedException, InvalidDataException
	{
		if (bioAssaySet == null) throw new InvalidUseOfNullException("bioAssaySet");
		bioAssaySet.checkPermission(Permission.WRITE);
		getData().setBioAssaySet(bioAssaySet.getData());
	}
	
	/**
		Get the type of the extra values.
		@return An <code>ExtraValueType</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the extra value type
		@throws BaseException If there is another error
	*/
	public ExtraValueType getExtraValueType()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(ExtraValueType.class, getData().getExtraValueType());
	}

	/**
		Set the extra value type. This cannot be changed after the extra value
		has been created.
	*/
	private void setExtraValueType(ExtraValueType extraValueType)
		throws PermissionDeniedException, InvalidDataException
	{
		if (extraValueType == null) throw new InvalidUseOfNullException("extraValueType");
		extraValueType.checkPermission(Permission.USE);
		getData().setExtraValueType(extraValueType.getData());
	}

	/**
		Get the job that created the extra values.
		@return A <code>Job</code> object, ot null if the extra values
			wasn't created by a job
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Job getJob()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Job.class, getData().getJob());
	}
	
	/**
		Set the job that created the extra values. Once set it can't 
		be changed.
	*/
	private void setJob(Job job)
	{
		getData().setJob(job == null ? null : job.getData());
	}
	
	/**
		Get the raw data type the experiment uses for the raw data.
		@return A <code>RawDataType</code> object
	*/
	public RawDataType getRawDataType()
	{
		return RawDataTypes.getSafeRawDataType(getData().getBioAssaySet().getExperiment().getRawDataType());
	}

	/**
		Get the number of extra values inserted.
	*/
	public int getNumValues()
	{
		return getData().getNumValues();
	}
	/**
		Called from batcher when it is closed.
	*/
	void setNumValues(int numValues)
	{
		getData().setNumValues(numValues);	
	}

	/**
		Get the number of values that are stored in files.
		@since 2.8
	*/
	public int getNumFileValues()
	{
		return getData().getNumFileValues();
	}
	
	/**
		Set the number of extra values that are stored in files.
		@param numValues The number of spots
		@since 2.8
	*/
	public void setNumFileValues(int numValues)
	{
		checkPermission(Permission.WRITE);
		getData().setNumFileValues(numValues);
	}
	
	/**
		Get the number of the data cube the extra values are stored in.
	*/
	public short getDataCubeNo()
	{
		return getData().getDataCubeExtraValue().getDataCube().getCubeNo();
	}
	
	/**
		Get the number in the data cube of the extra values.
	*/
	public short getExtraNo()
	{
		return getData().getDataCubeExtraValue().getExtraNo();
	}

	/**
		Get information about which coordinates in the data cube
		the extra values are attached to.
		@return A <code>CoordinateType</code> object
		@see DataCube
	*/
	public CoordinateType getCoordinateType()
	{
		return CoordinateType.fromValue(getData().getDataCubeExtraValue().getCoordinateType());	
	}
	
	/**
		Get the type of values. Shortcut for <code>getExtraValueType().getValueType()</code>.
		@return A <code>Type</code> object
	*/
	public Type getValueType()
	{
		return Type.fromValue(getData().getExtraValueType().getValueType());	
	}

	/**
		Get a query returning the extra values.
		The query will select column, position and the extra value by default,
		if the extra values attached to both column and position. Otherwise
		only the coordinate the extra values are attached to are selected.
<pre class="code">
// Example of default query for extra integer values attached to spots
SELECT ex.column, ex.position, ex.value
FROM Dynamic#SpotExtraInt AS ex
WHERE ex.cube = &lt;cube_no&gt; AND ex.extra = &lt;extra_no&gt;
</pre>

		@return A <code>DynamicExtraValueQuery</code> object
	*/
	public DynamicExtraValueQuery getValues()
	{
		VirtualColumn[] defaultSelects = new VirtualColumn[3];
		VirtualTable table = getVirtualTable();
		if (table.hasColumn(VirtualColumn.COLUMN)) defaultSelects[0] = VirtualColumn.COLUMN;
		if (table.hasColumn(VirtualColumn.POSITION)) defaultSelects[1] = VirtualColumn.POSITION;
		defaultSelects[2] = getVirtualColumn();
		
		DynamicExtraValueQuery query = new DynamicExtraValueQuery(this.getDataCube(), table, defaultSelects);
		query.restrictPermanent(
			new DynamicRestriction(VirtualColumn.EXTRA, getExtraNo())
		);
		return query;
	}

	DataCubeExtraValue getDataCubeExtraValue()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(DataCubeExtraValue.class, getData().getDataCubeExtraValue());
	}
	private void setDataCubeExtraValue(DataCubeExtraValue dcev)
	{
		if (dcev == null) throw new InvalidUseOfNullException("dcev");
		getData().setDataCubeExtraValue(dcev.getData());
	}
	
	VirtualDb getVirtualDb()
	{
		return getDbControl().getItem(VirtualDb.class, getData().getDataCubeExtraValue().getDataCube().getVirtualDb());
	}
	
	DataCube getDataCube()
	{
		return getDbControl().getItem(DataCube.class, getData().getDataCubeExtraValue().getDataCube());
	}

	VirtualTable getVirtualTable()
	{
		return getDataCubeExtraValue().getVirtualTable();
	}
	
	/**
		Get the virtual column object that represents the column where the
		extra values are stored. The column depends on the value type of the
		extra value.
		@return A <code>VirtualColumn</code> object
		@see VirtualColumn#extraValue(Type)
	*/
	public VirtualColumn getVirtualColumn()
	{
		return VirtualColumn.extraValue(getValueType());
	}
	
	/**
		The alias used when joining this extra value into a query. We can't use the
		<code>VirtualTable.getAlias()</code> value, which we normally do, since the alias 
		must be unique if we join several extra values. We simply return a value based on the
		id of this item. The {@link ExtraValueJoin} and {@link ExtraValueExpression}
		both uses this method to create their SQL.
		
		@return The alias to use in joins
		@see ExtraValueJoin
		@see ExtraValueExpression
	*/
	String getJoinAlias()
	{
		return "xtra"+getId();
	}
	
	/**
		Defines which coordinates in the data cube the extra values are attached to.
		Extra values can be attached to any combination of the column
		and postion coordinate.
		@see DataCube
	*/
	public enum CoordinateType
	{
		/**
			The extra values are attached to spots. Ie. position/column 
			coordinates.
		*/
		SPOT(0)
		{
			@Override
			public VirtualTable getVirtualTable(Type valueType)
			{
				if (valueType == Type.INT)
				{
					return VirtualTable.SPOT_EXTRA_INT;
				}
				else if (valueType == Type.FLOAT)
				{
					return VirtualTable.SPOT_EXTRA_FLOAT;
				}
				else if (valueType == Type.STRING)
				{
					return VirtualTable.SPOT_EXTRA_STRING;
				}
				else
				{
					throw new InvalidDataException("Value type is not supported as extra value: " + valueType);
				}
			}
		},
		
		/**
			The extra values are attached to position
			coordinates only.
		*/
		POSITION(1)
		{
			@Override
			public VirtualTable getVirtualTable(Type valueType)
			{
				if (valueType == Type.INT)
				{
					return VirtualTable.POSITION_EXTRA_INT;
				}
				else if (valueType == Type.FLOAT)
				{
					return VirtualTable.POSITION_EXTRA_FLOAT;
				}
				else if (valueType == Type.STRING)
				{
					return VirtualTable.POSITION_EXTRA_STRING;
				}
				else
				{
					throw new InvalidDataException("Value type is not supported as extra value: " + valueType);
				}
			}
		},

		
		/**
			The extra values are attached to column coordinates
			only.
		*/
		COLUMN(2)
		{
			@Override
			public VirtualTable getVirtualTable(Type valueType)
			{
				if (valueType == Type.INT)
				{
					return VirtualTable.COLUMN_EXTRA_INT;
				}
				else if (valueType == Type.FLOAT)
				{
					return VirtualTable.COLUMN_EXTRA_FLOAT;
				}
				else if (valueType == Type.STRING)
				{
					return VirtualTable.COLUMN_EXTRA_STRING;
				}
				else
				{
					throw new InvalidDataException("Value type is not supported as extra value: " + valueType);
				}
			}
		};
		
		/**
			Maps the integer that is stored in the database with an fileaction.
		*/
		private static final Map<Integer, CoordinateType> valueMapping = new HashMap<Integer, CoordinateType>();
		static
		{
			for (CoordinateType coordinateType : CoordinateType.values())
			{
				CoordinateType ct = valueMapping.put(coordinateType.getValue(), coordinateType);
				assert ct == null : "Another CoordinateType with the value "+coordinateType.getValue()+" already exists: " + ct;
			}
		}

		/**
			Get the <code>CoordinateType</code> object when you know the integer code.
			
			@param value The integer value
			@return The <code>CoordinateType</code> for the integer value,
				or null if not found
		*/
		public static CoordinateType fromValue(int value)
		{
			CoordinateType coordinateType = valueMapping.get(value);
			assert coordinateType != null : "coordinateType == null for value "+value;
			return coordinateType;
		}

		/**
			The integer value of this coordinate type.
			Used when stored in the database.
		*/
		private final int value;
		
		private CoordinateType(int value)
		{
			this.value = value;
		}
		
		/**
			Get the integer value that is used when storing a 
			<code>CoordinateType</code> in the database.
			@return The intger value of this coordinate type
		*/
		public int getValue()
		{
			return value;
		}
		
		/**
			Get the <code>VirtualTable</code> used to store extra values of
			this coordinate type given the value type.
			@param valueType The type of the values to store
		*/
		public abstract VirtualTable getVirtualTable(Type valueType);
	}
}
