/**
	$Id: FileStoreUtil.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import net.sf.basedb.core.data.DataFileTypeData;
import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;

/**
	Utility class working with {@link FileStoreEnabled} items. 

	@author nicklas
	@version 2.5
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (må, 22 maj 2017) $
*/
public class FileStoreUtil
{
	/**
		Check if a file of the given data file type is present in a 
		{@link FileStoreEnabled} item's file set. Optionally, check if the 
		file has been validated as a valid file.
	 	
	 	@param dc DbControl to use when accessing the database.		
		@param item The {@link FileStoreEnabled} item to check
		@param dataFileType The external ID of the {@link DataFileType} to check for
		@param requireValid TRUE if the file must exist and be valid, FALSE if
			it must only exist. This parameter is ignored if the data file type
			doesn't have a validator
		@return TRUE if the file exists, FALSE otherwise
	*/
	public static boolean hasDataFile(DbControl dc, FileStoreEnabled item, String dataFileType, boolean requireValid)
	{
		if (item == null || !item.hasFileSet()) return false;
		org.hibernate.query.Query<FileSetMemberData> query = HibernateUtil.getPredefinedQuery(
			dc.getHibernateSession(), "GET_FILESETMEMBER_FOR_DATAFILETYPE", FileSetMemberData.class);
		/*
			SELECT mbr 
			FROM FileSetMemberData mbr
			WHERE mbr.dataFileType.externalId = :externalId
			AND mbr.fileSet = :fileSet
		*/
		query.setParameter("externalId", dataFileType, TypeWrapper.H_STRING);
		query.setParameter("fileSet", item.getFileSet().getId(), TypeWrapper.H_INTEGER);
		FileSetMemberData member = HibernateUtil.loadData(query);
		// The file doesn't exists
		if (member == null) return false;

		// The file exists and we don't care about validity
		if (!requireValid) return true;

		// If we know for sure if the file is valid or not, return the status
		if (member.isValid() != null) return member.isValid();
		
		// If data file type has no validator we accept the file as valid
		DataFileType dft = dc.getItem(DataFileType.class, member.getDataFileType());
		return dft.hasActiveValidator(dc);
	}

	/**
		Get the file of the given data file type in the 
		{@link FileStoreEnabled} item's file set assuming that at most
		one file exists.
	 	
	 	@param dc DbControl used to access the database.
		@param item The {@link FileStoreEnabled} item to check
		@param dataFileType The external ID of the {@link DataFileType} to check for
		@param requireValid TRUE to only get the data file if is a valid or unvalidated
			file
		@return A file item, or null if no (valid) file of the given type exists
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the file
		@since 3.0
	*/
	public static File getDataFile(DbControl dc, FileStoreEnabled item, String dataFileType, boolean requireValid)
	{
		if (item != null && item.hasFileSet())
		{
			org.hibernate.query.Query<DataFileTypeData> query = HibernateUtil.getPredefinedQuery(
				dc.getHibernateSession(), "GET_DATAFILETYPE_FOR_EXTERNAL_ID", DataFileTypeData.class);
			query.setParameter("externalId", dataFileType, TypeWrapper.H_STRING);
			DataFileTypeData dft = HibernateUtil.loadData(query);
			if (dft != null)
			{
				int typeId = dft.getId();
				for (FileSetMemberData member : item.getFileSet().getData().getMembers())
				{
					if (member.getDataFileType().getId() == typeId)
					{
						if (!requireValid || !Boolean.FALSE.equals(member.isValid()))
						{
							return dc.getItem(File.class, member.getFile());
						}
					}
				}
			}
		}
		return null;
	}

	
	/**
		Get the file of the given data file type in the 
		{@link FileStoreEnabled} item's file set.
	 	
	 	@param dc DbControl used to access the database.
		@param item The {@link FileStoreEnabled} item to check
		@param dataFileType The external ID of the {@link DataFileType} to check for
		@param requireValid TRUE to only get the data file if is a valid or unvalidated
			file
		@return A list with file items, empty if no (valid) file of the given type exists
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the file
		@since 3.0
	*/
	public static List<File> getDataFiles(DbControl dc, FileStoreEnabled item, String dataFileType, boolean requireValid)
	{
		List<File> files = new ArrayList<File>();
		if (item != null && item.hasFileSet())
		{
			org.hibernate.query.Query<DataFileTypeData> query = HibernateUtil.getPredefinedQuery(
				dc.getHibernateSession(), "GET_DATAFILETYPE_FOR_EXTERNAL_ID", DataFileTypeData.class);
			query.setParameter("externalId", dataFileType, TypeWrapper.H_STRING);
			DataFileTypeData dft = HibernateUtil.loadData(query);
			if (dft != null)
			{
				int typeId = dft.getId();
				for (FileSetMemberData member : item.getFileSet().getData().getMembers())
				{
					if (member.getDataFileType().getId() == typeId)
					{
						if (!requireValid || !Boolean.FALSE.equals(member.isValid()))
						{
							files.add(dc.getItem(File.class, member.getFile()));
						}
					}
				}
			}
		}
		return files;
	}
	
	/**
		Get all files with a specific generic {@link ItemSubtype} in a given item's file
		set. Note! This method will only return files that the logged in user
		has permission to read. Other files in the file set will be ignored. 
		Note! This method checks the file type from {@link DataFileType#getGenericType()},
		not the file type from {@link File#getItemSubtype()}.
		
		@param dc A DbControl to use for database access
		@param item The {@link FileStoreEnabled} item to check
		@param genericType The system ID of the {@link ItemSubtype} to check for
		@return A list of file items that matches
	*/
	public static List<File> getGenericDataFiles(DbControl dc, FileStoreEnabled item, 
		String genericType)
	{
		List<File> files = new LinkedList<File>();
		if (item.hasFileSet())
		{
			ItemQuery<FileSetMember> query = item.getFileSet().getMembers(genericType);
			query.order(Orders.asc(Hql.property("file.name")));
			for (FileSetMember member : query.list(dc))
			{
				try
				{
					files.add(member.getFile());
				}
				catch (Throwable t)
				{}
			}
		}
		return files;
	}
	
	/**
		Get all file members with a specific generic {@link ItemSubtype} in a given item's file
		set. Note! This method will only return members that the logged in user
		has permission to read the file. Other members in the file set will be ignored. 
		Note! This method checks the file type from {@link DataFileType#getGenericType()},
		not the file type from {@link File#getItemSubtype()}.
		
		@param dc A DbControl to use for database access
		@param item The {@link FileStoreEnabled} item to check
		@param genericType The system ID of the {@link ItemSubtype} to check for
		@return A list of member items that matches
		@since 3.0
	*/
	public static List<FileSetMember> getGenericMembers(DbControl dc, FileStoreEnabled item, 
		String genericType)
	{
		List<FileSetMember> members = new LinkedList<FileSetMember>();
		if (item.hasFileSet())
		{
			ItemQuery<FileSetMember> query = item.getFileSet().getMembers(genericType);
			query.order(Orders.asc(Hql.property("file.name")));
			for (FileSetMember member : query.list(dc))
			{
				try
				{
					member.getFile();
					members.add(member);
				}
				catch (Throwable t)
				{}
			}
		}
		return members;
	}

	/**
		Get a list of all platform file types that has the specified generic {@link ItemSubtype}.
		If the item doesn't have any associated platform or variant, no file types are returned.
		
		@param dc A DbControl to use for database access
		@param item The {@link FileStoreEnabled} item to check
		@param genericType The system ID of the {@link ItemSubtype} to check for
		@return  A list of platform file types that matches
		@since 3.0
	*/
	public static List<PlatformFileType> getGenericPlatformFileTypes(DbControl dc, FileStoreEnabled item,
			String genericType)
	{
		Platform platform = item.getPlatform();
		PlatformVariant variant = item.getVariant();
		if (platform == null && variant == null) 
		{
			return Collections.emptyList();
		}
		
		ItemQuery<PlatformFileType> query = 
			PlatformFileType.getQuery(platform, variant, variant == null);
		query.restrict(
			Restrictions.eq(
				Hql.property("dataFileType.genericType.systemId"), 
				Expressions.string(genericType)
			)
		);
		query.order(Orders.asc(Hql.property("dataFileType.name")));
		return query.list(dc);
	}
	
	/**
		Set or remove a file to {@link FileStoreEnabled} item based on
		the file's {@link DataFileType}.

		@param dc A DbControl to use for database access
		@param item The item to add/remove the file to/from
		@param dataFileType The external ID of the {@link DataFileType} of the file
		@param file The file, or null to remove an existing file
		@return The new FileSetMember item or null if a file was removed
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission to the item
		@throws InvalidUseOfNullException If item is null
	*/
	public static FileSetMember setDataFile(DbControl dc, FileStoreEnabled item, String dataFileType, File file)
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		item.checkPermission(Permission.WRITE);
		FileSet fileSet = item.getFileSet();
		DataFileType type = DataFileType.getByExternalId(dc, dataFileType);
		FileSetMember member = fileSet.setMember(file, type);
		return member;
	}
	
	/**
		Set or remove a file to {@link FileStoreEnabled} item based
		on the file's generic {@link ItemSubtype}. Note! This method checks the file 
		type from {@link DataFileType#getGenericType()},
		not the file type from {@link File#getItemSubtype()}.
	
		@param dc A DbControl to use for database access
		@param item The item to add/remove the file to/from
		@param genericType The system ID of the generic {@link ItemSubtype}
		@param fallbackDataFileType The external ID of a {@link DataFileType} that is used
			if the specified generic file type isn't found among the file types
			registered for the platform
		@param file The file, or null to remove an existing file
		@return The new FileSetMember item or null if a file was removed
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission to the item
		@throws InvalidUseOfNullException If item is null
	*/
	public static FileSetMember setGenericDataFile(DbControl dc, FileStoreEnabled item, 
		String genericType, String fallbackDataFileType, File file)
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		item.checkPermission(Permission.WRITE);
		
		// Find out which DataFileType that is used as the generic type for the 
		// given platform/variant 
		Platform platform = item.getPlatform();
		PlatformVariant variant = item.getVariant();
		
		ItemQuery<PlatformFileType> query = 
			PlatformFileType.getQuery(platform, variant, variant == null);
		query.restrict(
			Restrictions.eq(
				Hql.property("dataFileType.genericType.systemId"), 
				Expressions.string(genericType)
			)
		);
		List<PlatformFileType> types = query.list(dc);
		DataFileType type = null;
		if (types.size() > 0) type = types.get(0).getDataFileType();
		if (type == null) type = DataFileType.getByExternalId(dc, fallbackDataFileType);
		
		// Now, add/remove the file to the fileset
		FileSet fileSet = item.getFileSet();
		FileSetMember member = fileSet.setMember(file, type);
		return member;
	}
	
}
