/*
	$Id $

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.KitData;
import net.sf.basedb.core.hibernate.TypeWrapper;

/**
	This class is used to represent a kit of reagents and other stuff used in 
	the lab process.
	
	@author Nicklas
	@since 3.7
*/
public class Kit
	extends AnnotatedItem
	implements Registered, Subtypable
{
	/**
		The type of item represented by this class.
		
		@see Item#KIT
		@see #getType()
	*/
	private static final Item TYPE = Item.KIT;

	/**
		The ID for the sampling kit type, for example a kit
		used when creating a {@link Sample} from a {@link
		BioSource}.
	*/
	public static final String SAMPLING = "net.sf.basedb.core.KitType.SAMPLING";

/**
		The ID for the extraction kit type, for example a kit
		used when creating an {@link Extract} from a {@link
		Sample}.
	*/
	public static final String EXTRACTION = "net.sf.basedb.core.KitType.EXTRACTION";

	/**
		The ID for the labeling kit type, for example a kit
		used when creating a labeled extract from an {@link Extract}.
	*/
	public static final String LABELING = "net.sf.basedb.core.KitType.LABELING";
	
	/**
		The ID for the library preparation kit type, for example a kit
		used when creating a library extract from an {@link Extract}.
	*/
	public static final String LIBRARY_PREPARATION = "net.sf.basedb.core.KitType.LIBRARY_PREPARATION";
	
	/**
		The ID for the hybridization kit type, for example a kit
		used when creating a {@link PhysicalBioAssay} from a set of
		(labeled) {@link Extract}:s.
	*/
	public static final String HYBRIDIZATION = "net.sf.basedb.core.KitType.HYBRIDIZATION";
	
	/**
		The ID for the cluster generation (in a flow cell) kit type, for example 
		a protocol used when creating a {@link PhysicalBioAssay}.
	*/
	public static final String CLUSTER_GENERATION = "net.sf.basedb.core.KitType.CLUSTER_GENERATION";

	/**
		The ID for the sequencing kit type, for example a kit
		used when creating a {@link DerivedBioAssay} from a {@link PhysicalBioAssay}.
	*/
	public static final String SEQUENCING = "net.sf.basedb.core.KitType.SEQUENCING";

	/**
		Get the system id for the kit subtype that should be used on an item
		that doesn't have a subtype. This includes {@link Subtypable} items with a
		null {@link ItemSubtype} and all other non-subtypable items which can be
		linked to a kit.
		@param itemType The item type
		@return The system id of the default kit type or null
	*/
	public static String getDefaultSystemId(Item itemType)
	{
		String defaultId = null;
		if (itemType == Item.SAMPLE)
		{
			defaultId = SAMPLING;
		}
		else if (itemType == Item.EXTRACT)
		{
			defaultId = EXTRACTION;
		}
		return defaultId;
	}

	/**
		Create a new <code>Kit</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@return The new <code>Kit</code> item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static Kit getNew(DbControl dc)
		throws BaseException
	{
		Kit kit = dc.newItem(Kit.class);
		kit.setName("New kit");
		kit.getData().setEntryDate(new Date());
		return kit;
	}

	/**
		Get a <code>Kit</code> item when you know the ID.
	
		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Kit</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Kit getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Kit kit = dc.loadItem(Kit.class, id);
		if (kit == null) throw new ItemNotFoundException("Kit[ID="+id+"]");
		return kit;
	}
	
	/**
		Get a query configured to retrieve kits.
		
		@return An {@link ItemQuery} object
	 	@throws BaseException If getting the query fails. 
	*/
	public static ItemQuery<Kit> getQuery() 
		throws BaseException
	{
		return new ItemQuery<Kit>(Kit.class);
	}
	
	/**
		Creates a new kit item.
		
		@param data the data
	*/
	Kit(KitData data)
	{
		super(data);
	}
	
	@Override
	KitData getData()
	{
		return (KitData)getBasicData();
	}
	/*
		From the Identibiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		@return Always null
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	@Override
	public void setItemSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype != null)
		{
			subtype.setOnItem(this);
		}
		else
		{
			getData().setItemSubtype(null);
		}
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>a {@link BioMaterialEvent} is linked to this kit
		<li>a {@link BioPlateEvent} is linked to this kit
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		boolean used = false;
		if (!used)
		{
			org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOMATERIALEVENTS_FOR_KIT", Long.class, "count(*)");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				WHERE bme.kit = :kit
			*/
			query.setParameter("kit", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		if (!used)
		{
			org.hibernate.query.Query<Long> q = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOPLATEEVENTS_FOR_KIT", Long.class, "count(*)");
				/*
					SELECT {1}
					FROM BioPlateEventData bpe
					WHERE bpe.kit = :kit
				*/
			q.setParameter("kit", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(q) > 0;
		}
		if (!used)
		{
			org.hibernate.query.Query<Long> q = HibernateUtil.getPredefinedQuery(session, 
					"GET_DERIVEDBIOASSAYS_FOR_KIT", Long.class, "count(*)");
					/*
						SELECT {1}
						FROM DerivedBioAssayData dba
						WHERE dba.kit = :kit
					*/
				q.setParameter("kit", getId(), TypeWrapper.H_INTEGER);
				used = HibernateUtil.loadData(q) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link BioMaterial}:s and {@link PhysicalBioAssay}:s using this kit
		<li>{@link BioPlateEvent}:s with this kit
		<li>{@link DerivedBioAssay}: with this kit
		</ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Biomaterial events - biomaterials
		org.hibernate.query.Query<BasicData> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_BIOMATERIALEVENTS_FOR_KIT", BasicData.class, "bme.bioMaterial");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				WHERE bme.kit = :kit
			*/
		query.setParameter("kit", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, query);

		// Biomaterial events - physical bioassays
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_BIOMATERIALEVENTS_FOR_KIT", BasicData.class, "bme.physicalBioAssay");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				WHERE bme.kit = :kit
			*/
		query.setParameter("kit", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, query);
		
		// Bioplate events
		org.hibernate.query.Query<Integer> query2 = HibernateUtil.getPredefinedQuery(session, 
			"GET_BIOPLATEEVENTS_FOR_KIT", Integer.class, "bpe.id");
			/*
				SELECT {1}
				FROM BioPlateEventData bpe
				WHERE bpe.kit = :kit
			*/
		query2.setParameter("kit", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.BIOPLATEEVENT, query2);
		
		// Derived bioassays
		query2 = HibernateUtil.getPredefinedQuery(session, 
			"GET_DERIVEDBIOASSAYS_FOR_KIT", Integer.class, "dba.id");
			/*
				SELECT {1}
				FROM DerivedBioAssayData dba
				WHERE dba.kit = :kit
			*/
		query2.setParameter("kit", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.DERIVEDBIOASSAY, query2);

		return using;
	}
	// -------------------------------------------


	/**
		Check if the kit is inactive. An inactive kit should not be used
		when creating new biomaterial. This is a recommendation only
		and is not enforced by the core API.
	*/
	public boolean isInactive()
	{
		return getData().isInactive();
	}
	
	/**
		Set the inactive flag. 
		
	 	@param inactive TRUE if the kit is inactive, FALSE otherwise. 
	 	@throws PermissionDeniedException If the logged in user doesn't 
	 		have write permission on the item.
	*/
	public void setInactive(boolean inactive)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setInactive(inactive);
	}

	/**
		Get the expiration date of the kit. A kit should not be used
		after it's expiration date. This is a recommendation only
		and is not enforced by the core API.
	*/
	public Date getExpirationDate()
	{
		return DateUtil.copy(getData().getExpirationDate());
	}

	/**
		Set the expiration date. 
		
	 	@param expirationDate A date or null
	 	@throws PermissionDeniedException If the logged in user doesn't 
	 		have write permission on the item.
	*/
	public void setExpirationDate(Date expirationDate)
	{
		checkPermission(Permission.WRITE);
		getData().setExpirationDate(DateUtil.setNullableDate(expirationDate, "expirationDate"));
	}
}
