/*
	$Id: Software.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.SoftwareData;
import net.sf.basedb.core.hibernate.TypeWrapper;

/**
	This class is used to represent individual software items
	and information about them. In the current version of BASE
	the only type of software we keep information about is feature extraction software.

	@author enell
	@version 2.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (må, 22 maj 2017) $
*/
public class Software
	extends AnnotatedItem
	implements Registered, Subtypable
{
	/**
		The type of item represented by this class.
		
		@see Item#SOFTWARE
		@see #getType()
		@base.modified $Date: 2017-05-22 14:35:27 +0200 (må, 22 maj 2017) $
	*/
	private static final Item TYPE = Item.SOFTWARE;

	/**
		The id for the <code>SoftwareType</code> item representing the
		softwaretype "feature extraction software".
	 	@since 3.0
	*/
	public static final String FEATURE_EXTRACTION = "net.sf.basedb.core.SoftwareType.FEATURE_EXTRACTION";

	/**
		The id for the <code>SoftwareType</code> item representing software
		used for aligning sequenced data against a known reference database.
	 	@since 3.0
	*/
	public static final String ALIGNMENT = "net.sf.basedb.core.SoftwareType.ALIGNMENT";

	
	/**
		The maximum length of the version variable that can be stored in the
		database. Check the length against this value before calling the
		{@link #setVersionString(String)} method to avoid
		exceptions.
	*/
	public static final int MAX_VERSIONSTRING_LENGTH = SoftwareData.MAX_VERSIONSTRING_LENGTH;
	
	/**
		Get the system id for the software subtype that should be used on an item
		that doesn't have a subtype. This includes {@link Subtypable} items with a
		null {@link ItemSubtype} and all other non-subtypable items which can be
		linked to a software.
		@param itemType The item type
		@return The system id of the default software type or null
		@since 3.0
	*/
	public static String getDefaultSystemId(Item itemType)
	{
		String defaultId = null;
		if (itemType == Item.RAWBIOASSAY)
		{
			defaultId = FEATURE_EXTRACTION;
		}
		return defaultId;
	}

	
	/**
		Create a new <code>Software</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@return The new <code>Software</code> item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static Software getNew(DbControl dc)
		throws BaseException
	{
		Software software = dc.newItem(Software.class);
		software.setName("New software");
		software.getData().setEntryDate(new Date());
		return software;
	}

	/**
		Get a <code>Software</code> item when you know the ID.
	
		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Software</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Software getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Software sw = dc.loadItem(Software.class, id);
		if (sw == null) throw new ItemNotFoundException("Software[ID="+id+"]");
		return sw;
	}
	
	/**
		Get a query configured to retrieve software.
		
		@return An {@link ItemQuery} object
	 	@throws BaseException If getting the query fails. 
	*/
	public static ItemQuery<Software> getQuery() 
		throws BaseException
	{
		return new ItemQuery<Software>(Software.class);
	}
	
	/**
		Creates a new software item.
		
		@param data the data
	*/
	Software(SoftwareData data)
	{
		super(data);
	}
	
	@Override
	SoftwareData getData()
	{
		return (SoftwareData)getBasicData();
	}
	/*
		From the Identibiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		@return Always null
		@since 3.3
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	@Override
	public void setItemSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype != null)
		{
			subtype.setOnItem(this);
		}
		else
		{
			getData().setItemSubtype(null);
		}
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>a RawBioAssay was created with this software
		<li>a DerivedBioassay was created with this software
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		boolean used = false;
		if (!used)
		{
			org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_DERIVEDBIOASSAYS_FOR_SOFTWARE", Long.class, "count(*)");
			/*
				SELECT {1}
				FROM DerivedBioAssayData dba
				WHERE dba.software = :software
			*/
			query.setParameter("software", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		if (!used)
		{
			org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_RAWBIOASSAYS_FOR_SOFTWARE", Long.class, "count(*)");
			/*
				SELECT {1}
				FROM RawBioAssay rba
				WHERE rba.software = :software
			*/
			query.setParameter("software", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link RawBioAssay}:s created with this software
		<li>{@link DerivedBioAssay}: created with this software
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_SOFTWARE", Integer.class, "rba.id");
			/*
				SELECT {1}
				FROM RawBioAssay rba
				WHERE rba.software = :software
			*/
		query.setParameter("software", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.RAWBIOASSAY, query);
		
		// Derived bioassay sets
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_DERIVEDBIOASSAYS_FOR_SOFTWARE", Integer.class, "dba.id");
			/*
				SELECT {1}
				FROM DerivedBioAssayData dba
				WHERE dba.software = :software
			*/
		query.setParameter("software", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.DERIVEDBIOASSAY, query);
		return using;
	}
	// -------------------------------------------

	/**
		Get the versionstring of this <code>Software</code> item.
		@return A string with the version of this item, or null if not known
	*/
	public String getVersionString()
	{
		return getData().getVersionString();
	}

	/**
		Set the versionstring for this <code>Software</code> item. The value
		must not be longer than the value specified by
		the {@link #MAX_VERSIONSTRING_LENGTH} constant.
		
		@param versionString The new version for this item, or null if not known
		@throws StringTooLongException If the version is longer then {@link #MAX_VERSIONSTRING_LENGTH}
		@throws PermissionDeniedException If the logged in user don't have write permission on this item
	*/
	public void setVersionString(String versionString) 
		throws StringTooLongException, PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setVersionString(StringUtil.setNullableString(versionString, "versionString", MAX_VERSIONSTRING_LENGTH));
	}
}
