/*
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.authentication;

import net.sf.basedb.core.data.UserData;

/**
	Objects of this class are returned by the 
	{@link AuthenticationManager#authenticate()}
	method and contains information about the authenticated
	user. This class provide information about the login and/or 
	external id for the user as well as the {@link AuthenticationMethod}
	that was used. Additional information for the user
	may be provided by {@link ExtraUserInformation}.
	
	@author Nicklas
	@since 3.3
*/
public class AuthenticatedUser
{

	private final AuthenticationMethod method;
	private final String login;
	private final String externalId;
	private final int internalId;
	private ExtraUserInformation extraInfo;
	
	/**
		Creates a new <code>AuthenticatedUser</code> object without any
		link to an existing user in BASE.
		
		@param method The authentication method used to verify the user
		@param login The login name
		@param externalId The unique externalId of the user
	*/
	public AuthenticatedUser(AuthenticationMethod method, String login, String externalId)
	{
		this.method = method;
		this.login = login;
		this.externalId = externalId;
		this.internalId = 0;
	}
	
	/**
		Creates a new <code>AuthenticatedUser</code> object with 
		a link to an existing user in BASE.
		
		@param method The authentication method used to verify the user
		@param user The UserData object for an existing user
	*/
	public AuthenticatedUser(AuthenticationMethod method, UserData user)
	{
		this.method = method;
		this.login = user.getLogin();
		this.externalId = user.getExternalId();
		this.internalId = user.getId();
	}

	/**
		Get the authentication method that was used to verify the login.
	*/
	public AuthenticationMethod getAuthenticationMethod()
	{
		return method;
	}
	
	/**
		Get the login string for the user.
	*/
	public String getLogin()
	{
		return login;
	}
	
	/**
		Get the external id for the user.
	*/
	public String getExternalId()
	{
		return externalId;
	}

	/**
		Get the internal id for the user (0 if not known)
	*/
	public int getInternalId()
	{
		return internalId;
	}

	/**
		Check if extra information about the user has been provided.
	*/
	public boolean hasExtraInformation()
	{
		return extraInfo != null;
	}
	
	/**
		Get extra information about the user.
	*/
	public ExtraUserInformation getExtraInformation()
	{
		return extraInfo;
	}
	
	/**
		Set extra information about the user.
	*/
	public void setExtraInformation(ExtraUserInformation extraInfo)
	{
		this.extraInfo = extraInfo;
	}
}
