/*
	$Id: GuiContext.java 6040 2012-03-30 08:22:05Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import net.sf.basedb.core.Item;

import java.util.Map;
import java.util.HashMap;

/**
	

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2012-03-30 10:22:05 +0200 (fr, 30 mar 2012) $
*/
public class GuiContext
{
	/**
		Shortcut to create a single-item context.
		@since 2.9
	*/
	public static GuiContext item(Item item)
	{
		return new GuiContext(item, Type.ITEM);
	}

	/**
		Shortcut to create a single-item context
		with a subcontext.
		@since 3.2
	*/
	public static GuiContext item(Item item, String subContext)
	{
		return new GuiContext(item, Type.ITEM, subContext);
	}

	
	/**
		Shortcut to create a list context.
		@since 2.9
	*/
	public static GuiContext list(Item item)
	{
		return new GuiContext(item, Type.LIST);
	}

	/**
		Shortcut to create a list context with a
		subcontext.
		@since 3.2
	*/
	public static GuiContext list(Item item, String subContext)
	{
		return new GuiContext(item, Type.LIST, subContext);
	}
	
	private final Item item;
	private final Type type;
	private final String subContext;
	
	/**
	 	@param item Item the context holds.
	 	@param type Type of context, 
	 		shows if it is a list or single view.
		
	*/
	public GuiContext(Item item, Type type)
	{
		this.item = item;
		this.type = type;
		this.subContext = null;
	}
	
	public GuiContext(Item item, Type type, String subContext)
	{
		this.item = item;
		this.type = type;
		this.subContext = subContext;
	}

	public Item getItem()
	{
		return item;
	}

	public Type getType()
	{
		return type;
	}

	public String getSubContext()
	{
		return subContext;
	}
	
	/**
		Checks if the specified object is an instance of the
		BasicItem subclass used by the {@link Item} in this context.
		@param o The object to check
		@since 2.9
	*/
	public boolean isCorrectBasicItem(Object o)
	{
		return item.getItemClass().isInstance(o);
	}
	
	/**
		Checks if the specified object is an instance of the
		BasicData subclass used by the {@link Item} in this context.
		@param o The object to check
		@since 2.9
	*/
	public boolean isCorrectBasicData(Object o)
	{
		return item.getDataClass().isInstance(o);
	}
	
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || other.getClass() != this.getClass()) return false;
		GuiContext gc = (GuiContext)other;
		return gc.item == this.item && gc.type == this.type;
	}
	
	@Override
	public int hashCode()
	{
		return item.hashCode() + type.hashCode();
	}
	
	@Override
	public String toString()
	{
		return item + (type == Type.ITEM ? " (item)" : " (list)");
	}

	public enum Type
	{
		ITEM(0),
		LIST(1);
		/**
			Maps the integer that is stored in the database.
		*/
		private static final Map<Integer, Type> valueMapping = new HashMap<Integer, Type>();
		
		static
		{
			for(Type type : Type.values())
			{
				Type t = valueMapping.put(type.getValue(), type);
				assert t == null : "Another GuiContext.Type with the value "+type.getValue()+" already exists";
			}
		}
		
		/**
			Get the <code>GuiContext.Type</code> object when you know the integer code.
			
			@param value The integer value.
			@return The type object or null if the value is unknown
		*/
		public static Type fromValue(int value)
		{
			Type type = valueMapping.get(value);
			assert type != null : "type == null for value "+value;
			return type;
		}

		private final int value;
		private Type(int value)
		{
			this.value = value;
		}
		
		/**
			Get the integer value that is used when storing a <code>GuiContext.Type</code> 
			to the database.
			@return The intger value of this type 
		*/
		public int getValue()
		{
			return value;
		}
	}
}
