/**
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.snapshot;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.units.UnitConverter;

/**
	Utility implementation for simple loading of primary annotations
	via {@link SnapshotManager}.

	@author Nicklas
	@since 3.4
*/
public class AnnotationLoaderUtil
	implements Filter<AnnotationSnapshot>
{
	private final DbControl dc;
	private final SnapshotManager manager;
	private final AnnotationType at;
	private final int atId;
	private final Type valueType;
	private final Unit defultUnit;
	private final boolean searchPrimary;
	private final boolean searchInherited;
	
	private Filter<AnnotationSnapshot> extraFilter;
	private List<AnnotationSnapshot> snapshots;
	private UnitConverter converter;
	private Unit unit;
	private String unitSymbol;
	
	/**
		Create a loder that uses the given snapshot manager to load annotations
		for a single annotation type. The loader will load only primary annotations.
	*/
	public AnnotationLoaderUtil(DbControl dc, SnapshotManager manager, AnnotationType at)
	{
		this(dc, manager, at, true, false);
	}
	
	/**
		Create a loder that uses the given snapshot manager to load annotations
		for a single annotation type. The loader can be configured to load primary 
		and/or inherited annotations.
		@since 3.5
	*/
	public AnnotationLoaderUtil(DbControl dc, SnapshotManager manager, AnnotationType at, boolean searchPrimary, boolean searchInherited)
	{
		this.dc = dc;
		this.manager = manager;
		this.at = at;
		this.searchPrimary = searchPrimary;
		this.searchInherited = searchInherited;
		this.atId = at.getId();
		this.valueType = at.getValueType();
		this.defultUnit = at.getDefaultUnit();
	}

	
	/*
		From the Filter interface
		-------------------------
	*/
	@Override
	public boolean evaluate(AnnotationSnapshot snapshot)
	{
		return snapshot.getAnnotationTypeId() == atId && (extraFilter == null || extraFilter.evaluate(snapshot));
	}
	// -------------------------

	/**
		The ID of the annotation type that this loader load annotations from.
	*/
	public int getId()
	{
		return atId;
	}
	
	/**
		The annotation type that this loader load annotations from.
	*/
	public AnnotationType getAnnotationType()
	{
		return at;
	}
	
	/**
		Is this loader searching for primary annotations?
		@since 3.5
	*/
	public boolean isSearchingPrimaryAnnotations()
	{
		return searchPrimary;
	}

	/**
		Is this loader searching for inherited annotations?
		@since 3.5
	*/
	public boolean isSearchingInheritedAnnotations()
	{
		return searchInherited;
	}

	
	/**
		Same as findAll but only return a boolean instead of the number
		of snapshots found.
		
		@return TRUE if an annotation was found, FALSE if not
		@see #findAll(AnnotationSetSnapshot)
	*/
	public boolean find(AnnotationSetSnapshot setSnapshot)
	{
		return findAll(setSnapshot, null) > 0;
	}
	
	/**
		Same as findAll but only return a boolean instead of the number
		of snapshots found.
		
		@return TRUE if an annotation was found, FALSE if not
		@see #findAll(AnnotationSetSnapshot, Filter)
		@since 3.11
	*/
	public boolean find(AnnotationSetSnapshot setSnapshot, Filter<AnnotationSnapshot> filter)
	{
		return findAll(setSnapshot, filter) > 0;
	}
	
	/**
		Find and load annotations from the given annotation set snapshot. 
		@return The number of snapshots found
		@since 3.5
	*/
	public int findAll(AnnotationSetSnapshot setSnapshot)
	{
		return findAll(setSnapshot, null);
	}
	
	/**
		Find and load annotations from the given annotation set snapshot
		using an optional extra filter.
		@return The number of snapshots found
		@since 3.11
	*/
	public int findAll(AnnotationSetSnapshot setSnapshot, Filter<AnnotationSnapshot> filter)
	{
		this.extraFilter = filter;
		snapshots = manager.findAnnotations(dc, setSnapshot, this, searchPrimary, searchInherited);
		int size = snapshots.size();
		if (size > 1)
		{
			unit = defultUnit;
		}
		else if (size == 1)
		{
			unit = snapshots.get(0).getActualUnit(dc);
		}
		else
		{
			unit = null;
		}
		if (unit != null)
		{
			converter = unit.equals(defultUnit) ? null : unit.getUnitConverter(defultUnit);
			unitSymbol = unit.getDisplaySymbol();
		}
		else
		{
			converter = null;
			unitSymbol = null;
		}
		return snapshots.size();
	}

	/**
		Get the snapshot for the last annotation found when calling
		{@link #findAll(AnnotationSetSnapshot)}. Note! If multiple 
		annotation snapshots were found, this will only return the 
		first hit.
		
		@return A snapshot or null if no annotations are found
	*/
	public AnnotationSnapshot getSnapshot()
	{
		return snapshots.size() > 0 ? snapshots.get(0) : null;
	}
	
	/**
		Get all snapshots found by the last {@link #findAll(AnnotationSetSnapshot)}
		call.
		@return A list with AnnotationSnapshot objects
	*/
	public List<AnnotationSnapshot> getSnapshots()
	{
		return snapshots;
	}
	
	/**
		Get the values from the last annotation found when calling
		{@link #find(AnnotationSetSnapshot)}
		@return A list of annotation values or null if no annotations are found
	*/
	public List<? extends Serializable> getValues()
	{
		List<Serializable> values = new ArrayList<Serializable>();
		for (AnnotationSnapshot s : snapshots)
		{
			values.addAll(s.getActualValues(converter, valueType));
		}
		return values;
	}
	
	/**
		Get the unit of the last annotation values that was found
		by {@link #find(AnnotationSetSnapshot)}. Note that if
		multiple annotations was found with different units the default
		unit of the annotation type is used.
		@return A string or null if the annotation doesn't use units
	*/
	public Unit getUnit()
	{
		return unit;
	}

	
	/**
		Get the unit symbol of the last annotation values that was found
		by {@link #find(AnnotationSetSnapshot)}.
		@return A string or null if the annotation doesn't use units
	*/
	public String getUnitSymbol()
	{
		return unitSymbol;
	}

	
}
