/*
	$Id: ClassUtil.java 7513 2018-11-02 07:45:02Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Set;

import net.sf.basedb.core.BaseException;

import java.util.HashSet;
import java.util.Collections;

/**
	Utility class for {@link Class} objects.


	@author Nicklas
	@version 2.0
	@base.modified $Date: 2018-11-02 08:45:02 +0100 (fr, 02 nov 2018) $
*/
public class ClassUtil
{

	/**
		Utility method for creating a new instance from a class with a default 
		no-argument constructor. Exceptions are wrapped in BaseExceptions.
		@since 3.14
	*/
	public static <I> I newInstance(Class<I> clazz)
	{
		try 
		{
			return clazz.getDeclaredConstructor().newInstance();
		} 
		catch (InstantiationException | IllegalAccessException | NoSuchMethodException | InvocationTargetException ex) 
		{
			throw new BaseException(ex);
		}	
	}
	
	/**
		Find a constructor for the specified class which takes parameters of
		the specified types taking superclasses and interfaces into
		account. This method finds the first declared constructor which can
		accept parameters of the specified type.
		
		@param clazz The class to find a constructor for
		@param parameters The parameters that are going to be sent to the 
			constructor's {@link Constructor#newInstance(Object[])} method
		@return A constructor object or null if no constructor was found
	*/
	@SuppressWarnings({"unchecked"})
	public static <I> Constructor<I> findConstructor(Class<I> clazz, Object[] parameters)
	{
		Constructor<I>[] all = (Constructor<I>[])clazz.getDeclaredConstructors();
		for (Constructor<I> c : all)
		{
			if (isAssignable(c.getParameterTypes(), parameters))
			{
				return c;
			}
		}
		return null;
	}
	
	/**
		Check if one array of class objects is assignment compatible with values in
		another array. Ie. the {@link Class#isAssignableFrom(Class)} method returns
		true for each element in the <code>classes</code> array with respect to the class
		of the corresponding value in the <code>values</code> array. 
		Null elements in the <code>values</code> array are ignored.
		@param classes The class objects representing the parameters we want to assign values to
		@param values The parameter values to assign
		@return TRUE if all parameters are assignable
	*/
	public static boolean isAssignable(Class<?>[] classes, Object[] values)
	{
		if (classes.length != values.length) return false;
		for (int i = 0; i < classes.length; ++i)
		{
			if (values[i] != null && !classes[i].isAssignableFrom(values[i].getClass())) return false;
		}
		return true;
	}
	
	/**
		Get all interfaces a class implements, including those implemented
		by superclasses.

		@param clazz The <code>Class</code> to check
		@return A <code>Set</code> containing the <code>Class</code> objects 
			for the interfaces
	*/
	public static Set<Class<?>> getInterfaces(Class<?> clazz)
	{
		Set<Class<?>> interfaces = new HashSet<Class<?>>();
		while (clazz != null)
		{
			Collections.addAll(interfaces, clazz.getInterfaces());
			clazz = clazz.getSuperclass();
		}
		return interfaces;
	}

	
	/**
		Check if a specified class exists and, optionally, if it has public no-argument
		constructor and implements a set of specific interfaces or superclasses.
		
		@param loader The classloader used to load the class, or null
			to use the default classloader
		@param name The fully qualified name of class to check
		@param publicNoArgConstructor TRUE to check if a public no-argument constructor exists
		@param interfaces An array of class objects representing interfaces or superclasses
			that the class to check must implement or extend
		@return The <code>Class</code> object for the named class
		@throws ClassNotFoundException If a class with the specified name isn't found
		@throws NoSuchMethodException If a public no-argument constructor isn't found
		@throws ClassCastException If the class doesn't implement or extend all of
			the interfaces/superclasses
		@since 2.5
	*/
	@SuppressWarnings("unchecked")
	public static <C> Class<? extends C> checkAndLoadClass(ClassLoader loader, String name,
		boolean publicNoArgConstructor, Class<C> iface, Class<?>... interfaces)
		throws ClassNotFoundException, NoSuchMethodException, ClassCastException
	{
		if (loader == null) loader = ClassUtil.class.getClassLoader();
		Class<?> clazz = loader.loadClass(name);
		checkClass(clazz, publicNoArgConstructor, iface);
		if (interfaces != null && interfaces.length > 0)
		{
			checkClass(clazz, publicNoArgConstructor, interfaces);
		}
		return (Class<? extends C>)clazz;
	}
	
	
	/**
		Check if a specified class has public no-argument
		constructor and implements a set of specific interfaces or superclasses.
		
		@param clazz The clazz to check
		@param publicNoArgConstructor TRUE to check if a public no-argument constructor exists
		@param interfaces An array of class objects representing interfaces or superclasses
			that the class to check must implement or extend
		@throws NoSuchMethodException If a public no-argument constructor isn't found
		@throws ClassCastException If the class doesn't implement or extend all of
			the interfaces/superclasses
		@since 2.10
	*/
	public static void checkClass(Class<?> clazz, 
		boolean publicNoArgConstructor, Class<?>... interfaces)
		throws NoSuchMethodException, ClassCastException
	{
		if (publicNoArgConstructor)
		{
			clazz.getConstructor((Class<?>[])null);
		}
		if (interfaces != null)
		{
			for (Class<?> iFace : interfaces)
			{
				if (iFace != null && !iFace.isAssignableFrom(clazz))
				{
					throw new ClassCastException("Class '" + clazz.getName() + "' is not a '" + iFace.getName() + "'");
				}
			}
		}
	}
	
	/**
		Find a method in a <code>clazz</code> with the given <code>name</code>
		that takes the given list of <code>arguments</code>. This method is
		just calling {@link Class#getMethod(String, Class...)} but returns
		null instead of throwing an exception in case the method is not found.

		@param clazz The class to look for the method in
		@param name The name of the method
		@param arguments The class types of the arguments
		@return The method or null if not found
		@since 2.7
	*/
	public static Method findMethod(Class<?> clazz, String name, Class<?>... arguments)
	{
		Method m = null;
		try
		{
			m = clazz.getMethod(name, arguments);
		}
		catch (NoSuchMethodException ex)
		{}
		return m;
	}

}
