/**
	$Id: DataParser.java 5319 2010-04-20 08:02:16Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.bfs;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.regex.Pattern;

import net.sf.basedb.core.File;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.encode.EncoderDecoder;
import net.sf.basedb.util.encode.TabCrLfEncoderDecoder;
import net.sf.basedb.util.parser.FlatFileParser;

/**
	Parser implementation that parses BFS data files. Before parsing 
	is started the stream to parse must be specified by calling
	{@link #setInputStream(InputStream)}. The actual parsing can be done
	in two different ways:
	<ul>
	<li>"Manual" parsing, using the {@link #parseToBof()} and {@link #nextData()} 
		methods.
	<li>Event-based parsing using the {@link #parse(EventHandler)} method.
		This parser issues the following events:
		<ul>
		<li>{@link #DATA_EVENT}: For each data line that is parsed.
		<li>{@link #END_OF_FILE_EVENT}: When the end of the file has been reached.
		</ul>
		
		The {@link MatrixModel} class implements a simple event handler that 
		collects the information from the data file and provides methods
		for accessing it, though it is recommended only for small files.
	</ul>
	
	<p>
	This class may be subclassed to provide customized behaviour.
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-04-20 10:02:16 +0200 (ti, 20 apr 2010) $
*/
public class DataParser
	implements BfsParser
{


	/**
		Event type that is issued for each data line. The event data is a
		string array with that data.
	*/
	public static final EventType<String[]> DATA_EVENT = 
		new EventType<String[]>("data.data");

	/**
		Event type that is issued when the end-of-file had been reached.
		No event data is submitted.
	*/
	public static final EventType<Object> END_OF_FILE_EVENT = 
		new EventType<Object>("data.end");
	
	/**
		Utility method for creating a data parser when you have an
		input stream.
		@param in The input stream the parser should read from
		@param filename Optional, the name of the file the input stream
			is reading from
	*/
	public static DataParser create(InputStream in, String filename, long fileSize)
	{
		DataParser parser = new DataParser();
		parser.setFilename(filename);
		parser.setInputStream(in);
		parser.setFileSize(fileSize);
		return parser;
	}
	
	/**
		Utility method for creating a data parser for a file in the BASE
		file system.
		@param file The file in the BASE file system
	*/
	public static DataParser create(File file)
	{
		return create(file.getDownloadStream(0), file.getName(), file.getSize());
	}
	
	/**
		Utility method for creating a data parser for a file in the local
		file system.
		@param file The file in the local file system
	*/
	public static DataParser create(java.io.File file)
		throws IOException
	{
		return create(new FileInputStream(file), file.getName(), file.length());
	}

	
	private final EncoderDecoder decoder;
	private InputStream in;
	private String filename;
	private long fileSize = -1;
	private FlatFileParser ffp;
	private int numColumns = -1;

	/**
		Create a new data parser.
	*/
	public DataParser()
	{
		this.decoder = new TabCrLfEncoderDecoder(true);
	}
	
	/**
		Decode an encoded value. Values are encoded/decoded
		with a {@link TabCrLfEncoderDecoder}.
		
		@param value The encoded value
		@return The decoded value
	*/
	public String decodeValue(String value)
	{
		return decoder.decode(value);
	}
	
	/**
		Get the file name that this parser is reading from.
		@return The file name or null if not known
	*/
	@Override
	public String getFilename()
	{
		return filename;
	}
	
	/**
		Set the file name that this parser is reading from.
	*/
	public void setFilename(String filename)
	{
		this.filename = filename;
	}
	
	/**
		Get the size in bytes of the file that this parser is reading from.
		@return The size or -1 if not known
	*/
	@Override
	public long getFileSize()
	{
		return fileSize;
	}
	
	/**
		Set the size of the file or -1 if not known.
	*/
	public void setFileSize(long fileSize)
	{
		this.fileSize = fileSize;
	}

	@Override
	public void close() 
	{
		if (in != null) FileUtil.close(in);
	}

	/**
		Set the input stream that should be parsed. This will also reset the
		parser. All information from a previously parsed file is lost.
		@param in The stream to parse (in UTF-8 format)
	*/
	public void setInputStream(InputStream in)
	{
		close();
		this.in = in;
		this.ffp = null;
		this.numColumns = -1;
	}
	
	/**
		Set the number of data columns that are expected to be found in this
		file.
		@param numColumns The number of columns or -1 to let the parser
			automatically detect this on the first data line
	*/
	public void setNumColumns(int numColumns)
	{
		this.numColumns = numColumns;
	}
	
	/**
		Parse the input stream and notify the specified event handler with events.
		The event handler is responsible for keeping track of and storing the
		data of interest.
		
		@param handler An event handler
		@throws IOException If there is an error reading the stream data
		@throws NullPointerException If the stream or event handler is null
	*/
	public void parse(EventHandler handler)
		throws IOException
	{
		if (handler == null) throw new NullPointerException("handler");
		parseToBof();
		
		// Repeat as long as there is data in the file
		String[] data = nextData();
		while (data != null)
		{
			ThreadSignalHandler.checkInterrupted();
			handleData(handler, data);
			data = nextData();
		}
		handleEndOfFile(handler);
	}
	
	/**
		Initialize parsing  of the data file. This method can only 
		be called once on  a given input stream. To parse a second file, call 
		{@link  #setInputStream(InputStream)} again to reset the parser.
		
		@throws IOException If the header line is not found
			 or if there is any other IO error
		@throws IllegalStateException If the parsing has already started
		@throws NullPointerException If no input stream has been specified
	*/
	public void parseToBof()
		throws IOException
	{
		if (in == null) throw new NullPointerException("inputStream");
		if (ffp != null) 
		{
			throw new IllegalStateException("Parsing of file '" + getFilename() + 
				"' has already started. Call setInputStream() to reset the parser.");
		}
		ffp = createFlatFileParser(in);
	}
	
	/**
		Get the next data line in the file.
		
		@return An array with the data, or null if no more data is found
		@throws IOException If there is an IO error or if the data is invalid
		@throws IllegalStateException If not {@link #parseToBof()} has been called
	*/
	public String[] nextData()
		throws IOException
	{
		if (ffp == null) 
		{
			throw new IllegalStateException("Parsing of file '" + getFilename() + 
				"' hasn't started. Call parseToBof() to begin parsing.");
		}
		if (!ffp.hasMoreData()) return null;

		String[] data = ffp.nextData().data();
		if (numColumns == -1)
		{
			numColumns = data.length;
		}
		else if (numColumns != data.length)
		{
			throw new IOException("Expected " + numColumns + " columns on line " + 
					ffp.getParsedLines() + " in file '" + getFilename() + 
					"'; found " + data.length + " columns");
		}
		
		// Decode the values
		for (int j = 1; j < numColumns; ++j)
		{
			data[j] = decodeValue(data[j]);
		}
		return data;
	}
	
	@Override
	public int getCurrentLine()
	{
		return ffp == null ? -1 : ffp.getParsedLines();
	}

	@Override
	public long getParsedBytes()
	{
		return ffp == null ? -1 : ffp.getParsedBytes();
	}
	
	/**
		Create a new flat file parser that can parse BFS annotation
		files. This method should set all regular expressions that
		are needed to parse the stream (assumed to be UTF-8).
	*/
	protected FlatFileParser createFlatFileParser(InputStream in)
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setDataSplitterRegexp(Pattern.compile("\\t"));
		ffp.setMinDataColumns(1);
		ffp.setMaxDataColumns(0); // = any number of columns
		ffp.setMaxUnknownLines(0);
		ffp.setInputStream(in, "UTF-8");
		return ffp;
	}
	
	/**
		Handle the data event. The default implemention sends
		a {@link #DATA_EVENT} to the event handler 
		@param handler The event handler from the {@link #parse(EventHandler)} method
		@param data An array with the data
	*/
	protected void handleData(EventHandler handler, String[] data)
	{
		handler.handleEvent(DATA_EVENT, data, this);
	}
	
	/**
		Handle the end-of-file event. The default implementation sends an
		{@link #END_OF_FILE_EVENT} notification to the event handler.
		@param handler The event handler from the {@link #parse(EventHandler)} method
	*/
	protected void handleEndOfFile(EventHandler handler)
	{
		handler.handleEvent(END_OF_FILE_EVENT, null, this);
	}

}
