/**
	$Id:XmlLoader.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.xml;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom2.Attribute;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.Namespace;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Presets.Preset;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XmlUtil2;
import net.sf.basedb.util.extensions.AboutBean;
import net.sf.basedb.util.extensions.Action;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.ErrorHandlerFactory;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionBean;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.ExtensionPointBean;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.RendererFactory;
import net.sf.basedb.util.filter.Filter;

/**
	Loads extension points and extensions from extension definition 
	XML files. This class may load extensions from more than one file
	and register or unregister all loaded extensions and extension points 
	with a registry.
	<p>
	For a description of the XML file format this class can load
	see the <a href="http://base.thep.lu.se/chrome/site/latest/html/developerdoc/extensions/extensions_developer.helloworld.html">BASE manual</a>.
	
	<p>
	<b>Factory initialisation</b><br>
	
	This loader uses reflection to create and initialise factory
	instances. The XML file contains the class name of the factory
	to create. This class must declare a public no-argument
	constructor and must also implement one of the specific
	factory interfaces (for example {@link ActionFactory}
	or {@link RendererFactory}). If the tag <code>&lt;parameters&gt;</code>
	is present inside the factory declaration, the loader will
	use reflection to find a setter method for each sub-tag inside
	the <code>&lt;parameters&gt;</code> tag. Here is an example:

	<pre class="code">
// XML contents
&lt;parameters&gt;
   &lt;image&gt;button.png&lt;/image&gt;
   &lt;title&gt;Click button&lt;/title&gt;
&lt;/parameters&gt;

// Factory initialization
factory.setImage("button.png");
factory.setTitle("Click button");
</pre>

	The setter methods must be public and accept a single <code>String</code> 
	parameter.
	<p>
	
	The loader also looks for the <code>setParameter(String, String)</code>
	method signature. If the method exists it will be called for each 
	parameter found with the tagname as the first parameter and the
	value as the second parameter. Continuing the above example:
	
	<pre class="code">
factory.setParameter("image", "button.png");
factory.setParameter("title", "Click button");
</pre>
	
	<p>
	Tags that doesn't have a corresponding setter method are simply ignored.
	The XML file format allows any tags as parameters, but only the first level
	will be parsed.
	
	<p>
	The parameter values may be subject to conversion by one or more 
	{@link ValueConverter}:s. Converters are added to the loader
	by {@link #addValueConverter(ValueConverter)} and are usually
	implemented to react on method annotations. For example, the web 
	client uses this for path and variable substitution.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class XmlLoader 
{
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(XmlLoader.class);

	/**
		The URL pointing to the extensions.xsd schema.
	*/
	public static final String SCHEMA_FILE_URL = 
		XmlLoader.class.getResource("/net/sf/basedb/core/xsd/extensions.xsd").toString();

	/**
		The name of the extensions namespace.
	*/
	public static final String NAMESPACE = "http://base.thep.lu.se/extensions.xsd";
	
	// To convert XML elements to a string representation
	private XMLOutputter xmlOut;
	
	// All loaded extension points
	private final List<ExtensionPoint<Action>> extensionPoints;
	
	// All loaded extensions
	private final List<Extension<Action>> extensions;
	
	// All loaded plug-in definitions
	private final List<PluginInfo> pluginDefinitions;
	
	// List of registered converters
	private Set<ValueConverter> converters;
	
	// Map with XML representation of factory parameters
	private final Map<Object, String> factoryParameters;
	
	// Map with classloaders
	private final Map<String, ClassLoader> classLoaders;
	
	// The last document that was validated
	private Document validatedDom;
	
	// The name of the last document
	private String lastName;
	
	/**
		Create a new XML loader instance.
	*/
	public XmlLoader()
	{
		extensionPoints = new LinkedList<ExtensionPoint<Action>>();
		extensions = new LinkedList<Extension<Action>>();
		pluginDefinitions = new LinkedList<PluginInfo>();
		factoryParameters = new HashMap<Object, String>();
		classLoaders = new HashMap<String, ClassLoader>();
		xmlOut = new XMLOutputter(Format.getPrettyFormat());
	}
	
	/**
		Add a value converter to this loader
		@param converter The converter to add
		@throws NullPointerException If the converter is null
	*/
	public void addValueConverter(ValueConverter converter)
	{
		if (converter == null) throw new InvalidUseOfNullException("converter");
		if (converters == null) converters = new LinkedHashSet<ValueConverter>();
		converters.add(converter);
	}

	/**
		Remove a converter.
		@param converter The converter to remove, null values are ignored
	*/
	public void removeValueConverter(ValueConverter converter)
	{
		if (converter == null || converters == null) return;
		converters.remove(converter);
	}
	
	/**
		Remove all value converters.
	*/
	public void clearValueConverters()
	{
		if (converters != null) converters.clear();
	}
	
	private Filter<Element> filter;
	/**
		Set a filter that all &lt;extension-point&gt;, &lt;extension&gt;,
		&lt;ref&gt; and &lt;plugin-definition&gt; tags must pass to be loaded 
		and registered.
		@param filter A filter or null to remove an existing filter
		@since 3.0
	*/
	public void setFilter(Filter<Element> filter)
	{
		log.debug("Using filter: " + filter);
		this.filter = filter;
	}
	
	/**
		Load an extensions definition XML file. This method may be called 
		multiple times with different XML files. The loading may also be
		done by a two-step process where the first step validates the file
		and the second step loads the extensions. See {@link 
		#validateXmlFile(InputStream, String)} and {@link 
		#loadLastValidatedFile(ClassLoader, boolean)}.

		@param xmlFile An input stream to read the XML data from
		@param filename The original filename the stream is coming from, or null
			if not known. This value is only used when generating error messages
		@param classLoader The classloader to use when loading classes that are
			named in the XML file, or null to use the default classloader (=the same
			class loader that loaded the BASE core classes)
		@param clear TRUE to clear all already loaded extensions before loading 
			the extensions in this file
		@return Information about the extension, or null if no such information
			is available
		@throws IOException If there is an error reading the XML file
		@throws ClassNotFoundException If a class named in the XML file can't
			be found
		@throws NoSuchMethodException If a factory class doesn't implement a
			public, no-argument constructor
		@throws IllegalAccessException If a factory class constructor isn't 
			public
		@throws InstantiationException If a factory class can't be instantiated,
			for example, because it is an abstract class or an interface
		@see #validateXmlFile(InputStream, String)
		@see #loadLastValidatedFile(ClassLoader, boolean)
	*/
	public About loadXmlFile(InputStream xmlFile, String filename,
			ClassLoader classLoader, boolean clear)
		throws IOException, ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException, InvocationTargetException
	{
		validateXmlFile(xmlFile, filename);
		return loadLastValidatedFile(classLoader, clear);
	}
	
	/**
		Validate an XML file against the extensions definition schema. If the
		file is valid you may continue to load the extensions. See
		{@link #loadLastValidatedFile(ClassLoader, boolean)}.
		
		@param xmlFile An input stream to read the XML data from
		@param filename The original filename the stream is coming from, or null
			if not known. This value is only used when generating error messages
		@return Information about the extension, or null if no such information
			is available
		@throws IOException If there is an error reading the XML file
	*/
	public About validateXmlFile(InputStream xmlFile, String filename)
		throws IOException
	{
		validatedDom = null;
		validatedDom = loadDocument(xmlFile, filename);
		lastName = filename;
		About globalAbout = loadGlobalAbout(validatedDom);
		verifyBaseVersion(globalAbout, true);
		return globalAbout;
	}
	
	/**
		Continue loading extensions from the last validated XML file. This is
		the second step in a two-step process and requires that 
		{@link #validateXmlFile(InputStream, String)} has been successfully called 
		first.

		@param classLoader The classloader to use when loading classes that are
			named in the XML file, or null to use the default classloader (=the same
			class loader that loaded the BASE core classes)
		@param clear TRUE to clear all already loaded extensions before loading 
			the extensions in this file
		@return Information about the extension, or null if no such information
			is available
		@throws ClassNotFoundException If a class named in the XML file can't
			be found
		@throws NoSuchMethodException If a factory class doesn't implement a
			public, no-argument constructor
		@throws IllegalAccessException If a factory class constructor isn't 
			public
		@throws InstantiationException If a factory class can't be instantiated,
			for example, because it is an abstract class or an interface
	*/
	public About loadLastValidatedFile(ClassLoader classLoader, boolean clear)
		throws ClassNotFoundException, IllegalAccessException, 
		InstantiationException, NoSuchMethodException, InvocationTargetException
	{
		if (validatedDom == null) 
		{
			throw new NullPointerException("No file has been validated");
		}
		if (clear)
		{
			extensionPoints.clear();
			extensions.clear();
			pluginDefinitions.clear();
		}
		About globalAbout = loadGlobalAbout(validatedDom);
		loadExtensionPoints(validatedDom, classLoader, globalAbout);
		loadExtensions(validatedDom, classLoader, globalAbout);
		loadPluginDefinitions(validatedDom, classLoader, globalAbout);
		validatedDom = null;
		return globalAbout;
	}
	
	/**
		Checks if an XML file has passed validation in the
		{@link #validateXmlFile(InputStream, String)} method. If so,
		the {@link #loadLastValidatedFile(ClassLoader, boolean)} can be called
		to continue loading the extensions.
		<p>
		Note that once the file has been loaded this flag is reset to
		FALSE.
		
		@return TRUE if a file has been validated, FALSE otherwise
	*/
	public boolean hasValidFile()
	{
		return validatedDom != null;
	}
	
	/**
		Get a list with all loaded extension points.
	*/
	public List<ExtensionPoint<Action>> getExtensionPoints()
	{
		return extensionPoints;
	}
	
	/**
		Get a list with all loaded extensions.
	*/
	public List<Extension<Action>> getExtensions()
	{
		return extensions;
	}
	
	/**
		Get a list with all loaded plug-in definitions.
		@since 3.0
	*/
	public List<PluginInfo> getPluginDefinitions()
	{
		return pluginDefinitions;
	}
	
	/**
		Get a XML-like string representation of the parameters that was
		used to initialise a factory.
		@param factory The factory object
		@return The parameters as found in the XML file
	*/
	public String getFactoryParameters(Object factory)
	{
		return factoryParameters.get(factory);
	}
	
	/**
		Register all loaded extension points with a registry.
		@param registry The registry
		@param update TRUE to update already registered extension
			points, FALSE to ignore
		@return The number of extension points registered or updated
		@see Registry#registerExtensionPoint(ExtensionPoint)
	*/
	public int registerExtensionPoints(Registry registry, boolean update)
	{
		int numRegistered = 0;
		for (ExtensionPoint<Action> ep : extensionPoints)
		{
			if (update || !registry.extensionPointIsRegistered(ep.getId()))
			{
				registry.registerExtensionPoint(ep, classLoaders.get(ep.getId()));
				numRegistered++;
			}
		}
		return numRegistered;
	}
	
	/**
		Unregister all loaded extension points from a registry.
		All extensions will also be unregistered.
		@param registry The registry
		@return The number of extension points that was unregistered
		@see Registry#unregisterExtensionPoint(String)
	*/
	public int unregisterExtensionPoints(Registry registry)
	{
		for (ExtensionPoint<Action> ep : extensionPoints)
		{
			registry.unregisterExtensionPoint(ep.getId());
		}
		return extensionPoints.size();
	}
	
	/**
		Register all loaded extensions with a registry.
		@param registry The registry
		@param update TRUE to update already registered extensions
			FALSE to ignore
		@return The number of registered or updated extensions
		@see Registry#registerExtension(Extension)
	*/
	public int registerExtensions(Registry registry, boolean update)
	{
		int numRegistered = 0;
		for (Extension<Action> ext : extensions)
		{
			if (update || !registry.extensionIsRegistered(ext.getId()))
			{
				registry.registerExtension(ext, classLoaders.get(ext.getId()));
				numRegistered++;
			}
		}
		return numRegistered;
	}
	
	/**
		Unregister all loaded extensions from a registry.
		@param registry The registry
		@return The number of extensions that was unregistered
		@see Registry#unregisterExtension(String)
	*/
	public int unregisterExtensions(Registry registry)
	{
		for (Extension<Action> ext : extensions)
		{
			registry.unregisterExtension(ext.getId());
		}
		return extensions.size();
	}
	
	/**
		Load and validate the XML file.
		@return A JDOM document object
	*/
	protected Document loadDocument(InputStream xmlFile, String filename)
		throws IOException
	{
		log.debug("Validating file: " + filename);
		return XmlUtil2.getSchemaValidatedXML(xmlFile, filename, NAMESPACE, SCHEMA_FILE_URL);
	}
	
	protected boolean verifyBaseVersion(About about, boolean throwException)
	{
		if (about == null) return true;
		String minVersion = about.getMinBaseVersion();
		String maxVersion = about.getMaxBaseVersion();
		if (minVersion != null && Version.compareWith(minVersion) > 0)
		{
			if (throwException)
			{
				throw new BaseException("BASE version "+ minVersion + " or higher is required");
			}
			return false;
		}
		if (maxVersion != null && Version.compareWith(maxVersion) <= 0)
		{
			if (throwException)
			{
				throw new BaseException("BASE version "+ maxVersion + " or higher is not supported");
			}
			return false;
		}
		return true;
	}
	
	/**
		Load the global about tag from the document.
		@return An about bean or null if the about tag is not found.
	*/
	protected AboutBean loadGlobalAbout(Document dom)
	{
		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		return loadAbout(root.getChild("about", ns));
	}
	
	/**
		Load information in an <code>&lt;about&gt;</code> tag.
		@param aboutTag The about tag
	*/
	protected AboutBean loadAbout(Element aboutTag)
	{
		AboutBean about = null;
		if (aboutTag != null)
		{
			Namespace ns = aboutTag.getNamespace();
			about = new AboutBean();
			about.setName(Values.getStringOrNull(aboutTag.getChildText("name", ns)));
			about.setDescription(Values.getStringOrNull(aboutTag.getChildText("description", ns)));
			about.setVersion(Values.getStringOrNull(aboutTag.getChildText("version", ns)));
			about.setMinBaseVersion(Values.getStringOrNull(aboutTag.getChildText("min-base-version", ns)));
			about.setMaxBaseVersion(Values.getStringOrNull(aboutTag.getChildText("max-base-version", ns)));
			about.setContact(Values.getStringOrNull(aboutTag.getChildText("contact", ns)));
			about.setCopyright(Values.getStringOrNull(aboutTag.getChildText("copyright", ns)));
			about.setEmail(Values.getStringOrNull(aboutTag.getChildText("email", ns)));
			about.setUrl(Values.getStringOrNull(aboutTag.getChildText("url", ns)));
			
			List<Attribute> attributes = aboutTag.getAttributes();
			for (Attribute a : attributes)
			{
				about.setAttribute(a.getName(), a.getValue());
			}
		}
		return about;
	}
	
	
	/**
		Create extension points from an XML document.
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	protected int loadExtensionPoints(Document dom, ClassLoader classLoader, About globalAbout)
		throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException, InvocationTargetException
	{
		log.debug("Loading extension points from file: " + lastName);
		int numLoaded = 0;
		List<ExtensionPoint<Action>> temp = new LinkedList<ExtensionPoint<Action>>();
		/*
			Format of an extension point is:
			<extension-point
				id="..."
				>
				<name>...</name>
				<description>...</description>
				<action-class>...</action-class>
				<renderer-factory override="true|false">
					<factory-class>...</factory-class>
					<parameters>
						any-tags
					</parameters>
				</renderer-factory>
				<error-handler-factory>
					<factory-class>...</factory-class>
					<parameters>
						any-tags
					</parameters>
				</error-handler-factory>
			</extension-point>
		 */
		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		// ID base will be prepended to extension point id:s if defined
		String idBase = Values.getStringOrNull(root.getAttributeValue("id-base"));
		
		List<Element> epTags = root.getChildren("extension-point", ns);
		log.debug("Found " + epTags.size() + " extension point(s) in file: " + lastName);
		for (Element epTag : epTags)
		{
			String id = Values.getStringOrNull(epTag.getAttributeValue("id"));
			if (idBase != null) id = idBase + id;

			log.debug("Processing extension point: " + id);
			
			if (filter == null || filter.evaluate(epTag))
			{
				ExtensionPointBean<Action> extensionPoint = new ExtensionPointBean<Action>();
				temp.add(extensionPoint);
				numLoaded++;
		
				// Set ID (required) and classloader
				extensionPoint.setId(id);
				
				// Set Name and Description (optional)
				extensionPoint.setName(Values.getStringOrNull(epTag.getChildText("name", ns)));
				extensionPoint.setDescription(Values.getStringOrNull(epTag.getChildText("description", ns)));
				
				// Load and set Action class (required)
				String actionClassName = Values.getStringOrNull(epTag.getChildText("action-class", ns));
				Class actionClass = 
					ClassUtil.checkAndLoadClass(classLoader, actionClassName, false, Action.class);
				extensionPoint.setActionClass(actionClass);
				
				// Load and set RendererFactory (optional)
				Element rfTag = epTag.getChild("renderer-factory", ns);
				if (rfTag != null)
				{
					RendererFactory<? super Action> rf = createFactory(rfTag, classLoader, RendererFactory.class, null);
					extensionPoint.setRendererFactory(rf);
					extensionPoint.setAllowRendererOverrider(
						Values.getBoolean(rfTag.getAttributeValue("override")));
				}
				
				Element ehfTag = epTag.getChild("error-handler-factory", ns);
				if (ehfTag != null)
				{
					ErrorHandlerFactory<? super Action> ehf = createFactory(ehfTag, classLoader, ErrorHandlerFactory.class, null);
					extensionPoint.setErrorHandlerFactory(ehf);
				}
				
				if (classLoader != null) classLoaders.put(extensionPoint.getId(), classLoader);
			}
		}
		extensionPoints.addAll(temp);
		return numLoaded;
	}
	

	/**
		Create extensions from an XML document.
	*/
	@SuppressWarnings("unchecked")
	protected int loadExtensions(Document dom, ClassLoader classLoader, About globalAbout)
		throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException, InvocationTargetException
	{
		log.debug("Loading extensions from file: " + lastName);
		int numLoaded = 0;
		List<Extension<Action>> temp = new LinkedList<Extension<Action>>();
		/*
			Format of an extension is:
			<extension
				id="..."
				extends="..."
				>
				<index>...</index>
				<about>
					<name>...</name>
					<description>...</name>
					<version>...</version>
					<contact>...</contact>
					<copyright>...</copyright>
					<email>...</email>
					<url>...</url>
				</about>
				<action-factory>
					<factory-class>...</factory-class>
					<parameters>
						any-tags
					</parameters>
				</action-factory>
				<action-factory>
					<factory-class>...</factory-class>
					<parameters>
						any-tags
					</parameters>
				</action-factory>
			</extension>
		*/

		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		// ID base will be prepended to extension point id:s if defined
		String idBase = Values.getStringOrNull(root.getAttributeValue("id-base"));

		List<Element> epTags = root.getChildren("extension", ns);
		log.debug("Found " + epTags.size() + " extension(s) in file: " + lastName);
		for (Element epTag : epTags)
		{
			// id
			String id = Values.getStringOrNull(epTag.getAttributeValue("id"));
			if (idBase != null) id = idBase + id;
			log.debug("Processing extension: " + id);
			
			// <index>
			float defaultIndex = Values.getFloat(epTag.getChildText("index", ns), 999.0f);
			
			// Pick up the extension points + index
			List<String> extensionPoints = new ArrayList<String>();
			List<Float> indexes = new ArrayList<Float>();
			// First... the extends attribute
			String epId = Values.getStringOrNull(epTag.getAttributeValue("extends"));
			if (epId != null)
			{
				if (filter == null || filter.evaluate(epTag))
				{
					extensionPoints.add(epId);
					indexes.add(defaultIndex);
				}
			}
			
			// Then... any <extends>/<ref> subtags
			Element extendsTag = epTag.getChild("extends", ns);
			if (extendsTag != null)
			{
				for (Element refTag : extendsTag.getChildren("ref", ns))
				{
					if (filter == null || filter.evaluate(refTag))
					{
						epId = Values.getStringOrNull(refTag.getText());
						if (epId != null)
						{
							extensionPoints.add(epId);
							indexes.add(Values.getFloat(refTag.getAttributeValue("index"), defaultIndex));
						}
					}
				}
			}

			// Any extension points remaining after filtering?
			if (extensionPoints.size() > 0)
			{
	
				// <about>
				AboutBean about = loadAbout(epTag.getChild("about", ns));
				if (globalAbout != null)
				{
					if (about == null) about = new AboutBean();
					about.copy(globalAbout, false);
				}
				
				if (!verifyBaseVersion(about, false))
				{
					String min = about.getMinBaseVersion();
					String max = about.getMaxBaseVersion();
					log.info("Extension '" + id + "' require BASE version between " + 
						(min == null ? "*" : min) + " and " + (max == null ? "*" : max)	
					);
				}
				else
				{
					// <action-factory>
					Element afTag = epTag.getChild("action-factory", ns);
					ActionFactory<? super Action> af = null;
					Preset config = Application.getExtensionsManager().getSettings().getSettingsForExtension(id);
					if (afTag != null)
					{
						af = createFactory(afTag, classLoader, ActionFactory.class, config);
					}
				
					// <renderer-factory>
					Element rfTag = epTag.getChild("renderer-factory", ns);
					RendererFactory<? super Action> rf = null;
					if (rfTag != null)
					{
						rf = createFactory(rfTag, classLoader, RendererFactory.class, config);
					}
					
					String commonIdPrefix = null;
					if (extensionPoints.size() > 1)
					{
						// We need to calculate the common id prefix for the extension points
						// so that we can create unique extension id:s
						commonIdPrefix = StringUtil.getCommonPrefix(extensionPoints);
					}
					
					for (int i = 0; i < extensionPoints.size(); ++i)
					{
						epId = extensionPoints.get(i);
						float index = indexes.get(i);
						ExtensionBean<Action> ext = new ExtensionBean<Action>();
						temp.add(ext);
						numLoaded++;
						
						String thisId = id;
						if (commonIdPrefix != null)
						{
							thisId += ":" + epId.substring(commonIdPrefix.length());
						}
						
						ext.setId(thisId);
						ext.setExtends(epId);
						ext.setIndex(index);
						ext.setAbout(about);
						ext.setActionFactory(af);
						ext.setRendererFactory(rf);
						if (classLoader != null) classLoaders.put(ext.getId(), classLoader);
					}
				}
			}
		}
		extensions.addAll(temp);
		return numLoaded;
	}
	
	/**
		Load plug-in information from an XML document.
		@since 3.0
	*/
	protected int loadPluginDefinitions(Document dom, ClassLoader classLoader, About globalAbout)
		throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException
	{
		log.debug("Loading plug-in definitions from file: " + lastName);
		int numLoaded = 0;
		List<PluginInfo> temp = new LinkedList<PluginInfo>();
		/*
			Format of a plug-in definition is:
			<plugin-definition
				id="..."
				>
				<about>
					<name>...</name>
					<description>...</name>
					<version>...</version>
					<contact>...</contact>
					<copyright>...</copyright>
					<email>...</email>
					<url>...</url>
				</about>
				<plugin-class>...</plugin-class>
			</plugin-definition>
		*/
	
		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		// ID base will be prepended to extension point id:s if defined
		String idBase = Values.getStringOrNull(root.getAttributeValue("id-base"));
	
		List<Element> pdTags = root.getChildren("plugin-definition", ns);
		log.debug("Found " + pdTags.size() + " plug-in definitions in file: " + lastName);
		for (Element pdTag : pdTags)
		{
			// id
			String id = Values.getStringOrNull(pdTag.getAttributeValue("id"));
			if (idBase != null) id = idBase + id;
			log.debug("Processing plug-in definition: " + id);
			
			if (filter == null || filter.evaluate(pdTag))
			{
				// <about>
				AboutBean about = loadAbout(pdTag.getChild("about", ns));
				if (globalAbout != null)
				{
					if (about == null) about = new AboutBean();
					about.copy(globalAbout, false);
				}
				
				if (!verifyBaseVersion(about, false))
				{
					String min = about.getMinBaseVersion();
					String max = about.getMaxBaseVersion();
					log.info("Plug-in definition '" + id + "' require BASE version between " + 
						(min == null ? "*" : min) + " and " + (max == null ? "*" : max)	
					);
				}
				else
				{
					// <plugin-class>
					String className = Values.getStringOrNull(pdTag.getChildText("plugin-class", ns));
					
					PluginInfo info = new PluginInfo(id);
					info.setAbout(about);
					info.setClassName(className);
					
					temp.add(info);
					
					// <settings>
					Element settingsTag = pdTag.getChild("settings", ns);
					if (settingsTag != null)
					{
						for (Element propertyTag : settingsTag.getChildren("property", ns))
						{
							String name = propertyTag.getAttributeValue("name");
							String value = Values.getStringOrNull(propertyTag.getText());
							info.setProperty(name, value);
						}
					}
				}
			}
		}
		pluginDefinitions.addAll(temp);
		return numLoaded;
	}

	
	/**
		Create a new factory instance. The factory must be a class
		with a public, no-argument constructor. The factory class name
		is read from the &lt;factory-class&gt; tag and initialisation
		parameters from the &lt;parameters&gt; tag.
		
		@param factoryTag The root tag of the factory definition
		@param classLoader The classloader to use, or null to use
			the BASE core classloader
		@param factoryType Class type of the factory. The named class
			must implement or be a subclass of this class
		@return An initialised factory
	*/
	protected <F> F createFactory(Element factoryTag, ClassLoader classLoader, Class<F> factoryType, Preset config)
		throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException, InvocationTargetException
	{
		Namespace ns = factoryTag.getNamespace();
		String factoryClassName = Values.getStringOrNull(factoryTag.getChildText("factory-class", ns));
		Class<?> factoryClass = 
			ClassUtil.checkAndLoadClass(classLoader, factoryClassName, true, factoryType);
		F factory = factoryType.cast(factoryClass.getDeclaredConstructor().newInstance());
		initBeanWithReflection(factory, factoryTag.getChild("parameters", ns), config);
		return factory;
	}
	
	/**
		Initialise a bean using reflection. For each child tag
		of the 'root' element, this method will check if the
		bean implements a setter method that is compatible with the
		child tag name. The method name should start with 'set' and
		then the tag name with the first letter capitalized. The method
		must take a single String argument. For example, if there is
		a child tag <code>&lt;image&gt;button.png&lt;/image&gt;</code> this will
		be converted to the method call: <code>setImage("button.png")</code>.
		<p>
		Tags that has no matching public setter method are ignored.
		
		@param bean The bean to initialize
		@param root The root element, if null nothing is done
	*/
	protected void initBeanWithReflection(Object bean, Element root, Preset config)
	{
		if (root == null) return;
		Class<?> beanClass = bean.getClass();
		
		List<Element> parameters = root.getChildren();
		factoryParameters.put(bean, xmlOut.outputString(root));
		
		// Check for generic setParameter(String, String) method
		Method setParameter = getSetParameterMethod(beanClass);
		
		// List all child tags
		for (Element child : parameters)
		{
			// childName is the name of the setter method --> set<ChildName>
			String childName = child.getName();

			// The value passed as an argument to the setter method
			String value = config != null ? Values.getStringOrNull(config.getSetting(childName)) : null;
			if (value == null) 
			{
				value = Values.getStringOrNull(child.getText());
			}
			
			if (value == null)
			{
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}

			Method setter = getSetterMethod(beanClass, childName);
			if (setter != null)
			{
				// Specific setter method was found
				// Convert the input value
				String converted = getConvertedValue(value, setter);

				// Invoke the setter method -- ignore all errors
				try
				{
					setter.invoke(bean, converted);
				}
				catch (IllegalAccessException ex)
				{}
				catch (InvocationTargetException ex)
				{}
			}
			else if (setParameter != null) 
			{
				// Call generic setParameter(String, String) method
				String converted = getConvertedValue(value, setParameter);
				try
				{
					setParameter.invoke(bean, childName, converted);
				}
				catch (IllegalAccessException ex)
				{}
				catch (InvocationTargetException ex)
				{}
			}
		}
	}
	
	/**
		Get the setter method for a given tag name.
		@param beanClass The class to look for the setter method in
		@param tagName The tag name in the XML file
		@return A Method object, or null if no method is found
	*/
	protected Method getSetterMethod(Class<?> beanClass, String tagName)
	{
		String methodName = getSetterMethodNameFromTag(tagName);
		Method setter = null;
		try
		{
			setter = beanClass.getMethod(methodName, String.class);
		}
		catch (NoSuchMethodException ex)
		{}
		return setter;
	}
	
	/**
		Check if the bean class has a <code>setParameter(String, String)</code>
		method and return it's reference if it has.
		@param beanClass The class to look for the method in
		@return A Method object, or null if no method is found
	*/
	protected Method getSetParameterMethod(Class<?> beanClass)
	{
		Method setParameter = null;
		try
		{
			setParameter = beanClass.getMethod("setParameter", String.class, String.class);
		}
		catch (NoSuchMethodException ex)
		{}
		return setParameter;
		
	}
	
	/**
		Convert the tag name to a setter method name. The tag name is prefixed with
		'set', the first letter is capitalized, hyphens in the tag name are removed and
		the first letter after each is capitalized.
		@param tagName The tag name
		@return The converted tag name
	*/
	protected String getSetterMethodNameFromTag(String tagName)
	{
		String[] parts = tagName.split("\\-");
		StringBuilder setterName = new StringBuilder(tagName.length()+3);
		setterName.append("set");
		for (String part : parts)
		{
			setterName.append(part.substring(0,  1).toUpperCase());
			setterName.append(part.substring(1));
		}
		return setterName.toString();
	}

	protected String getConvertedValue(String original, Method method)
	{
		String converted = original;
		if (converters != null)
		{
			for (ValueConverter converter : converters)
			{
				converted = converter.convert(converted, method);
			}
		}
		return converted;
	}
	
}
