/**
	$Id: BaseFileImporter.java 6077 2012-08-03 10:53:22Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DatabaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.MappingBatcher;
import net.sf.basedb.core.PositionBatcher;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SimpleAbsoluteProgressReporter;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.VirtualTable;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.ChainedProgressReporter;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.basefile.BaseFileParser;
import net.sf.basedb.util.importer.FileWrapper;
import net.sf.basedb.util.importer.spotdata.BaseFileInfo.ChildBioAssay;
import net.sf.basedb.util.parser.FlatFileParser;

/**
	Imports spot data from a serial or matrix BASEfile. Before it
	can be used various configuration properties must be set.
	<ul>
	<li>An open DbControl: {@link #setDbControl(DbControl)}
	<li>The source file: {@link #setSourceFileWrapper(FileWrapper)}
	<li>The destination transaction: {@link #setTransformation(Transformation)}
	</ul>
	
	Optional are:
	<ul>
	<li>A progress reporter: {@link #setProgressReporter(ProgressReporter)}
	<li>The intensity transform used by the data set: 
		{@link #setIntensityTransform(IntensityTransform)}.
		If not set, the same intensity transform as for the parent
		bioassay set is used
	<li>Redefine column names. By default, column names used by BASE 1 
		are assumed. Call {@link #useBase2ColumnNames(int)} to configure
		the parser to use BASE 2 column names.
	</ul>
	
	<p>
	The importer will create a new child bioassay set. Child bioassays are 
	created according to the data found in the BASEfile. The importer will
	create a new datacube if the reporter/position mapping has changed or
	if the child/parent bioassay relationship is not 1:1. The importer
	supports extra values (float) and will create missing extra value types
	if needed.
	
	@author Nicklas
	@version 2.14
	@base.modified $Date: 2012-08-03 12:53:22 +0200 (fr, 03 aug 2012) $
*/
public class BaseFileImporter
{

	private final BaseFileParser parser;
	private DbControl dc;
	private ProgressReporter progress;

	private FileWrapper sourceFile;
	private Transformation transformation;
	private IntensityTransform transform;

	
	/**
		Create a new importer object.
	*/
	public BaseFileImporter()
	{
		this.parser = new BaseFileParser();
	}
	
	/*
		public Configuration properties
		-------------------------------
	*/
	/**
		Set's the DbControl that should be used to get data from the database.
	*/
	public void setDbControl(DbControl dc)
	{
		this.dc = dc;
	}
	
	/**
		Get the current DbControl.
	*/
	public DbControl getDbControl()
	{
		return dc;
	}

	/**
		Set the progress reporter that is used to report progress. Call
		{@link #setProgress(int, String)} to update the current status.
	*/
	public void setProgressReporter(ProgressReporter progress)
	{
		this.progress = progress;
	}
	
	/**
		Get the progress reporter.
	*/
	public ProgressReporter getProgressReporter()
	{
		return progress;
	}
	
	/**
		Set the source BASEfile to import data from.
		@param source The file
		@since 2.16
	*/
	public void setSourceFileWrapper(FileWrapper source)
	{
		this.sourceFile = source;
	}
	
	/**
		Get the source file.
	*/
	public FileWrapper getSourceFileWrapper()
	{
		return sourceFile;
	}
		
	/**
		Set the destination transformation. A child bioassay set is created.
		@param transformation The transformation
	*/
	public void setTransformation(Transformation transformation)
	{
		this.transformation = transformation;
	}
	
	/**
		Get the destination transformation.
	*/
	public Transformation getTransformation()
	{
		return transformation;
	}
	
	/**
		Set the intensity transform that have been applied to the spot
		data in the file. If not specified, the same transform as for
		the parent bioassay set is used.
		@param transform The intensity transform
	*/
	public void setIntensityTransform(IntensityTransform transform)
	{
		this.transform = transform;
	}
	
	/**
		Get the intensity transform
	*/
	public IntensityTransform getIntensityTransform()
	{
		return transform;
	}
	
	/**
		Get the underlying parser for the BASEfile. It is useful when
		there is need to change for example, name of columns, 
		add different spot intensity parsers, etc.
		@return The BASEfile parser
	*/
	public BaseFileParser getBaseFileParser()
	{
		return parser;
	}
	
	/**
		Reconfigure the parser to use BASE 2 column names instead of BASE 1
		column names.
		@param channels The expected number of data channels in the file
		@since 2.16
	*/
	public void useBase2ColumnNames(int channels)
	{
		parser.setRedefinedColumnName("spots", "position", "Position");
		parser.setRedefinedColumnName("spots", "reporter", "Internal id");
		for (int ch = 1; ch <= channels; ++ch)
		{
			parser.setRedefinedColumnName("spots", "intensity" + ch, "Ch " + ch);
		}
	}
	// -------------------------------
	
	/*
		Helper methods
		--------------
	*/
	/**
		Update the progress of the export.
		@see ProgressReporter
	*/
	protected void setProgress(int percent, String message)
	{
		if (progress != null)
		{
			progress.display(percent, message);
		}
	}
	// ----------------------------

	/**
		Start the import.
		@return The created bioassay set or null if the BASEfile didn't contain
			spot data
	*/
	public BioAssaySet doImport()
		throws IOException
	{
		// Check required configuration parameters
		FileWrapper srcFile = getSourceFileWrapper();
		DbControl dc = getDbControl();
		Transformation t = getTransformation();
		if (srcFile == null) throw new InvalidUseOfNullException("sourceFile");
		if (dc == null) throw new InvalidUseOfNullException("dbControl");
		if (t == null) throw new InvalidUseOfNullException("transformation");
		
		// Get optional and derived parameters
		BioAssaySet parent = t.getSource();
		IntensityTransform transform = getIntensityTransform();
		if (transform == null) transform = parent.getIntensityTransform();
		ChainedProgressReporter chainedProgress = progress == null ? 
			null : new ChainedProgressReporter(progress);
		BaseFileInfo info = new BaseFileInfo(srcFile);
		
		// First parser pass: progress=0--30%
		SectionAssaysParser assaysParser = new SectionAssaysParser(dc, info);
		FirstPassSectionSpotsParser spotParser = new FirstPassSectionSpotsParser(dc, info, parent);
		SectionReporterListParser listParser = new SectionReporterListParser(dc, info, t);
		parser.setSectionParser("assays", assaysParser);
		parser.setSectionParser("spots", spotParser);
		parser.setSectionParser("reporterlist", listParser);
		if (chainedProgress != null)
		{
			chainedProgress.setRange(0, 30);
			parser.setProgressReporter(new SimpleAbsoluteProgressReporter(chainedProgress, srcFile.getSize()));
		}
		InputStream srcIn = srcFile.getInputStream();
		try
		{
			FlatFileParser ffp = parser.parse(srcIn, srcFile.getCharacterSet());
			int totalLines = ffp.getParsedLines();
			
			// If the BASEfile didn't have any 'spots' sections we don't 
			// create a child bioassay set
			if (parser.getSectionCount("spots") == 0) return null;
	
			// Create the child bioassay set and bioassays
			boolean useNewDataCube = info.getAssaysHaveParentAssaysMapping() 
				|| info.getChildHasDifferentReporterPositionMapping();
			BioAssaySet child = createChildBioAssaySet(dc, info, t, useNewDataCube);
			child.setIntensityTransform(transform);
			createChildBioAssays(dc, child, info, useNewDataCube);
			
			// Create the position/reporter mapping if needed
			if (useNewDataCube)
			{
				if (chainedProgress != null) chainedProgress.setRange(30, 40);
				createChildPositionReporterMapping(child, info, chainedProgress);
				if (!info.getChildHasDifferentReporterPositionMapping() && child.getRawDataType().isStoredInDb())
				{
					// If the child and parent has identical position/reporter mapping
					// we can "calculate" new raw data mappings.
					if (chainedProgress != null) chainedProgress.setRange(40, 50);
					createChildRawDataMapping(child, parent, info, chainedProgress);
				}
			}
			FileUtil.close(srcIn);
			
			// Second parser pass: progress=50-100%
			SecondPassSectionSpotsParser spotParser2 = 
				new SecondPassSectionSpotsParser(dc, info, child, totalLines);
			BaseFileParser secondParser = new BaseFileParser();
			secondParser.copyRedefinedColumnNames(parser);
			secondParser.setSectionParser("spots", spotParser2);
			if (chainedProgress != null)
			{
				chainedProgress.setRange(50, 100);
				secondParser.setProgressReporter(new SimpleAbsoluteProgressReporter(chainedProgress, totalLines));
			}
			srcIn = srcFile.getInputStream();
			secondParser.parse(srcIn, srcFile.getCharacterSet());
			return child;
		}
		finally
		{
			FileUtil.close(srcIn);
		}
	}

	/**
		Helper method for creating a child bioassay set using the information in the
		BaseFileInfo object. The created bioassay set is scheduled for saving to
		the database.
		
		@param dc A DbControl to use for database access
		@param info Information from the parsed BASEfile
		@param t The parent transformation
		@param useNewDataCube TRUE to create the child bioassay set in a new
			datacube, FALSE to use the same as the parent bioassay set
		@return The new child bioassay set
	*/
	public BioAssaySet createChildBioAssaySet(DbControl dc, BaseFileInfo info, 
		Transformation t, boolean useNewDataCube)
	{
		BioAssaySet child = null;
		if (useNewDataCube)
		{
			child = t.newProduct("new", "new", false);
		}
		else
		{
			child = t.newProduct(null, "new", false);
		}
		dc.saveItem(child);
		return child;
	}
	
	
	/**
		Helper method for creating new child bioassays using the information 
		found in the BaseFileInfo object.
		@see BaseFileInfo#getChildAssays()
	*/
	public void createChildBioAssays(DbControl dc, BioAssaySet childSet, 
			BaseFileInfo info, boolean useNewDataCube)
	{
		for (ChildBioAssay childData : info.getChildAssays())
		{
			BioAssay childBa = null;
			if (useNewDataCube)
			{
				Collection<Integer> parentIds = childData.getParents();
				Collection<BioAssay> parents = null;
				if (parentIds == null || parentIds.size() == 0)
				{
					parents = Collections.singletonList(BioAssay.getById(dc, childData.getFileId()));
				}
				else
				{
					parents = new ArrayList<BioAssay>(parentIds.size());
					for (Integer parentId : parentIds)
					{
						parents.add(BioAssay.getById(dc, parentId));
					}
				}
				childBa = childSet.newBioAssay(parents);

				// Store column mappings for future use
				short childColumn = childBa.getDataCubeColumnNo();
				for (BioAssay parentBa : parents)
				{
					info.mapParentChildColumns(parentBa.getDataCubeColumnNo(), childColumn);
				}
			}
			else
			{
				BioAssay parentBa = BioAssay.getById(dc, childData.getFileId());
				childBa = childSet.newBioAssay(parentBa);
				// Columns should be identical... put in the mapping just be be sure
				info.mapParentChildColumns(parentBa.getDataCubeColumnNo(), childBa.getDataCubeColumnNo());
			}
			if (childData.getName() != null) childBa.setName(childData.getName());
			dc.saveItem(childBa);
			childData.setChild(childBa);
		}
	}

	/**
		Helper method for creating position/reporter mappings for the child 
		bioassay set (which must be in a new datacube). 
		@param child The new child bioassay set
		@param info Information about the position/reporter mapping
			is found in {@link BaseFileInfo#getChildReporterPositions()}
		@param progress Optional progress reporter
	*/
	public void createChildPositionReporterMapping(BioAssaySet child, BaseFileInfo info, 
			ProgressReporter progress)
	{
		PositionBatcher batcher = child.getPositionBatcher();
		ReporterProxy proxy = new ReporterProxy();
		Map<Integer, Integer> mapping = info.getChildReporterPositions();
		int total = mapping.size();
		int done = 0;
		int delta = Math.max(total / 20, 50);
		
		for (Map.Entry<Integer, Integer> entry : mapping.entrySet())
		{
			ThreadSignalHandler.checkInterrupted();
			if (progress != null && done % delta == 0)
			{
				int percent = (100 * done) / total;
				progress.display(percent, "Mapping new reporter positions; " + 
						done + " of " + total);
			}
			int position = entry.getKey();
			Integer reporterId = entry.getValue();
			if (reporterId == null)
			{
				batcher.insert(position, null);
			}
			else
			{
				proxy.setTheId(reporterId);
				batcher.insert(position, proxy);
			}
			++done;
		}
		batcher.flush();
		batcher.close();
	}
	
	/**
		Helper method for creating a rawdata mapping for the child bioassay set.
		This can only be done if the child is in a new data cube and if the
		position/reporter mapping hasn't change. Since we know the child/parent
		bioassay set mapping, the raw data mapping for a spot in the child is the 
		same as the raw data mapping for the spots at the same position among the
		parents.
		<p>
		Child/parent mapping is taken from the BaseFileInfo object. See
		{@link BaseFileInfo#getChildColumns(short)}. This mapping automatically
		created by {@link #createChildBioAssays(DbControl, BioAssaySet, BaseFileInfo, boolean)}.
		
		@param child The new child bioassay set
		@param parent The parent bioassay set
		@param info Information about the child/parent mapping
			is found in {@link BaseFileInfo#getChildColumns(short)} 
		@param progress Optional progress reporter
	*/
	public void createChildRawDataMapping(BioAssaySet child, BioAssaySet parent, 
		BaseFileInfo info, ProgressReporter progress)
	{
		MappingBatcher mapBatcher = child.getMappingBatcher();
		DynamicSpotQuery spotQuery = parent.getSpotData();
		spotQuery.joinRawData(JoinType.INNER);
		spotQuery.select(Dynamic.select(VirtualColumn.COLUMN));
		spotQuery.select(Dynamic.select(VirtualColumn.POSITION));
		spotQuery.select(Dynamic.select(VirtualTable.RAWPARENTS, VirtualColumn.RAWDATA_ID));
		
		int total = parent.getNumSpots();
		int done = 0;
		int delta = Math.max(total / 20, 50);
		
		DynamicResultIterator spotIterator = spotQuery.iterate(dc);
		RawDataProxy rawProxy = new RawDataProxy();
		try
		{
			while (spotIterator.hasNext())
			{
				ThreadSignalHandler.checkInterrupted();
				if (progress != null && done % delta == 0)
				{
					int percent = (100 * done) / total;
					progress.display(percent, "Mapping new raw data positions; " + 
							done + " of " + total);
				}
				SqlResult result = spotIterator.next();
				short parentColumn = result.getShort(1);
				int position = result.getInt(2);
				int rawDataId = result.getInt(3);
				List<Short> childColumns = info.getChildColumns(parentColumn);
				if (childColumns != null)
				{
					rawProxy.setTheId(rawDataId);
					for (short childColumn : childColumns)
					{
						mapBatcher.insert(childColumn, position, rawProxy);
					}
				}
				++done;
			}
		}
		catch (SQLException ex)
		{
			throw new DatabaseException(ex);
		}
		mapBatcher.flush();
		mapBatcher.close();
	}
	
	static class ReporterProxy
		extends ReporterData
	{
		private int id;
		ReporterProxy()
		{}
		public void setTheId(int id)
		{
			this.id = id;
		}
		@Override
		public int getId()
		{
			return id;
		}
	}

	static class RawDataProxy
		extends RawData
	{
		private int id;
		RawDataProxy()
		{}
		public void setTheId(int id)
		{
			this.id = id;
		}
		@Override
		public int getId()
		{
			return id;
		}
	}	
	
}
