/**
	$Id: JobAgentServerConnection.java 7551 2018-12-12 14:41:10Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jobagent;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.nio.channels.ClosedByInterruptException;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;

import org.slf4j.Logger;

import net.sf.basedb.util.SocketUtil;

/**
	This class is used by job agents to listen for incoming
	requests from client applications. 
	
	@author nicklas
	@version 2.0
	@base.modified $Date: 2018-12-12 15:41:10 +0100 (on, 12 dec 2018) $
*/
public class JobAgentServerConnection
	implements AutoCloseable
{
	private final int port;
	private final RequestHandler requestHandler;
	private final Logger logger;
	private Thread listener;
	
	/**
		Create a new server connection. The {@link #open()} method must be called
		to start listening for incoming connections.
		
		@param port The port to listing to for incoming connections
		@param requestHandler A handler for taking care of incoming requests.
			The handler must be thread-safe and able to handle multiple 
			requests at the same time
		@param logger A logger object for logging debug and other information
			or null if no logging is wanted
	*/
	public JobAgentServerConnection(int port, RequestHandler requestHandler, Logger logger)
	{
		this.port = port;
		this.requestHandler = requestHandler;
		this.logger = logger;
	}
	
	/**
		Start listening for incoming connections. This method starts a new listener 
		thread that answers incoming requests. For each request a new thread
		is created and control is passed on to the {@link RequestHandler}. If a listener
		has already been create thi method does nothing.
		@throws IOException If there is an error
	*/
	public void open()
		throws IOException
	{
		if (listener != null) return;
		if (logger != null) logger.info("Opening listener on port " + port);
		
		ServerSocketChannel channel = ServerSocketChannel.open();
		channel.socket().bind(new InetSocketAddress(port));
		
		listener = new Thread(new ListenerThread(channel, requestHandler, logger), 
				"ListenerThread."+this.toString());
		listener.start();
		if (logger != null) logger.info("Now listening on port " + port);
	}
	
	/**
		Stop listening for incoming connections. The thread that listens for incoming 
		connections is killed. Any executing requests are not affected by this method.
	*/
	@Override
	public void close()
	{
		if (listener == null) return;
		if (logger != null) logger.info("Stopping listener on port " + port);
		listener.interrupt();
		listener = null;
	}

	/**
	 	Is a listener active or not for this server.
		@return TRUE if a listener is active, FALSE otherwise
	*/
	public boolean isListening()
	{
		return listener != null;
	}
	
	/**
		This class is used for listening to the specified socket for
		incoming connections. Each request is forwarded to the 
		<code>RequestHandler</code> in a separate thread.
	*/
	private static class ListenerThread
		implements Runnable
	{
		
		private final ServerSocketChannel socket;
		private final RequestHandler requestHandler;
		private final Logger logger;
		
		private ListenerThread(ServerSocketChannel socket, RequestHandler requestHandler, Logger logger)
		{
			this.socket = socket;
			this.requestHandler = requestHandler;
			this.logger = logger;
		}
		
		/*
			From the Runnable interface
			-------------------------------------------
		*/
		@Override
		public void run()
		{
			// We listen until someone tells us not to!
			boolean interrupted = false;
			while (!interrupted)
			{
				try
				{
					SocketChannel incoming = socket.accept();
					if (logger != null && logger.isDebugEnabled()) 
					{
						logger.debug("Accepted incoming connection from: " + 
							incoming.socket().getInetAddress().toString());
					}
					Thread request = new Thread(new RequestHandlerThread(incoming.socket(), 
							requestHandler, logger), "RequestHandlerThread." + this.toString());
					request.start();
					interrupted = Thread.interrupted();
				}
				catch (ClosedByInterruptException ex)
				{
					if (logger != null) logger.info("Listener service was interrupted by another thread");
					interrupted = true;
				}
				catch (Throwable t)
				{
					if (logger != null) logger.error(t.getMessage(), t);
					interrupted = true;
				}
			}
			if (logger != null) logger.info("Closing socket on port " + socket.socket().getLocalPort());
			SocketUtil.close(socket);
		}
		// -------------------------------------------
	}
	
	/**
		This class is used for forwarding requests to a {@link RequestHandler}.
		For each accepted request a new thread is created befor the request
		is forwarded to the <code>RequestHandler</code>.
	*/
	private static class RequestHandlerThread
		implements Runnable
	{
		private final Socket incoming;
		private final RequestHandler requestHandler;
		private final Logger logger;
		
		private RequestHandlerThread(Socket incoming, RequestHandler requestHandler, Logger logger)
		{
			this.incoming = incoming;
			this.requestHandler = requestHandler;
			this.logger = logger;
		}
		
		/*
			From the Runnable interface
			-------------------------------------------
		*/
		@Override
		public void run()
		{
			try
			{
				String cmd = SocketUtil.read(incoming, true);
				String answer = null;
				try
				{
					answer = requestHandler.handleCmd(incoming, cmd);
				}
				catch (Throwable t)
				{
					answer = t.getMessage() + " [" + t.getClass().getName() + "]";
					if (logger != null) logger.error("Eception in request handler "+ requestHandler + 
						": " + t.getMessage(), t);
				}
				SocketUtil.send(incoming, answer, true);
				incoming.close();
			}
			catch (Throwable t)
			{
				if (logger != null) logger.error(t.getMessage(), t);
			}
		}
		// -------------------------------------------
		
	}

	
	
}
