/**
	$Id: ArrayBatchLoader.java 5500 2010-11-18 07:26:30Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for array batches. The forward-loading
	direction goes from {@link ArrayDesign}:s -&gt; {@link ArrayBatch}:s -&gt; 
	{@link ArraySlide}:s. The reverse-loading direction is the opposite 
	direction.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2010-11-18 08:26:30 +0100 (to, 18 nov 2010) $
*/
public class ArrayBatchLoader
	extends BasicItemNodeLoader<ArrayBatch>
{

	public ArrayBatchLoader()
	{
		super(Item.ARRAYBATCH, ALLOW_ROOT_NODE, 
			new NameableNameGenerator<ArrayBatch>("batch", "Array batch"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading array batch nodes from a given array design node.
		The returned node is a folder-type node containing item nodes for 
		each batch that has been created from the array design.
		@return The folder node for the array batches
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node designNode)
	{
		NodeFactory<ArrayBatch> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ArrayDesign design = (ArrayDesign)designNode.getItem(dc);
		ItemQuery<ArrayBatch> query = context.initQuery(design.getArrayBatches(), "name");
		ItemResultIterator<ArrayBatch> it = query.iterate(dc);
		while (it.hasNext())
		{
			ArrayBatch batch = it.next();
			if (folderNode == null)
			{
				folderNode = new Node("batches", "Batches", designNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, batch, batch, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, designNode, false);
		return folderNode;
	}
	
	/**
		Create a reverse-loading array batch node from an array slide node.
		@return An array batch node, or null if the slide doesn't belong to a batch
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node slideNode)
	{
		NodeFactory<ArrayBatch> nf = getNodeFactory(dc, context);
		ArrayBatch batch = null;
		boolean denied = false;
		try
		{
			batch = ((ArraySlide)slideNode.getItem(dc)).getArrayBatch();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node batchNode = createItemNode(nf, batch, batch, denied, 
				slideNode, ChildNodeDirection.REVERSE);
		return batchNode;
	}
	// --------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of an array batch. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Print robot: {@link HardwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node batchNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, batchNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, batchNode);
		getNodeLoader(context, Item.HARDWARE).createPropertyNode(dc, context, batchNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, batchNode);
	}

	/**
		Loads the array slide nodes that is related to the given batch node.
		@see ArraySlideLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node batchNode)
	{
		getNodeLoader(context, Item.ARRAYSLIDE).createForwardNode(dc, context, batchNode);
	}
	
	/**
		Loads the array design node that is the parent of this array batch. 
		@see ArrayDesignLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node batchNode)
	{
		getNodeLoader(context, Item.ARRAYDESIGN).createReverseNode(dc, context, batchNode);
	}
	// ---------------------------------------
	
}
