/**
	$Id: BasicItemNodeLoaderFactory.java 7513 2018-11-02 07:45:02Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.overview.OverviewContext;

/**
	Node loader factory implementation for {@link BasicItem} node loaders.
	The key used in {@link #createNodeLoader(Object)} should normally be 
	an {@link Item} object.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2018-11-02 08:45:02 +0100 (fr, 02 nov 2018) $
*/
public class BasicItemNodeLoaderFactory
	implements NodeLoaderFactory<BasicItem, Object>
{
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(BasicItemNodeLoaderFactory.class);
	private static final boolean debug = log.isDebugEnabled();
	
	private final Map<Object, Class<? extends NodeLoader<? extends BasicItem>>> loaders;
	private boolean autoLoadChildren;
	private boolean useNullLoader;
	
	/**
		Creates a new factory.
	*/
	public BasicItemNodeLoaderFactory()
	{
		loaders = new HashMap<Object, Class<? extends NodeLoader<? extends BasicItem>>>();
		this.autoLoadChildren = false;
		this.useNullLoader = true;
		registerDefaultNodeLoaders();
	}
	
	/*
		From the NodeLoaderFactory interface
		------------------------------------
	*/
	/**
		Create a new node loader for loading nodes for the specified item
		type. If no loader has been registered for the given item type,
		a {@link NullNodeLoader} is returned or an exception is thrown
		depending on the {@link #useNullLoaderIfNotFound()} option.
		
		@param key An {@link Item} object
		@throws ItemNotFoundException If no node loader is found and the 
			{@link #useNullLoaderIfNotFound()} setting is false
		@return A node loader object
	*/
	@Override
	public NodeLoader<? extends BasicItem> createNodeLoader(Object key)
	{
		if (debug) log.debug("Create: NodeLoader[key=" + key + "]");
		NodeLoader<? extends BasicItem> loader = null;
		Class<? extends NodeLoader<? extends BasicItem>> loaderClass = loaders.get(key);
		if (loaderClass == null)
		{
			if (debug) log.debug("Not found: NodeLoader[key=" + key + "]");
			if (useNullLoader)
			{
				loader = new NullNodeLoader<BasicItem>();
			}
			else
			{
				throw new ItemNotFoundException("Node loader for key: " + key);
			}
		}
		else
		{
			if (debug) log.debug("Instantiating: " + loaderClass);
			try
			{
				loader = loaderClass.getDeclaredConstructor().newInstance();
			}
			catch (InstantiationException | IllegalAccessException | NoSuchMethodException | InvocationTargetException ex)
			{
				throw new BaseException("Can't create node loader: " + loaderClass, ex);
			}
		}
		// Wrap in extension loader
		loader = new ExtensionChildNodeLoader<>(loader);
		// ...and recursive child loader if auto-loading is enabled
		if (isAutoLoadingChildren()) loader = new AutoChildNodeLoader<>(loader);
		
		if (debug) log.debug("Node loader created: " + loader);
		return loader;
	}
	// --------------------------------------
	
	/**
		Register a node loader implementation. The implementation must have a
		public no-argument constructor. The given class is checked with
		{@link ClassUtil#checkAndLoadClass(ClassLoader, String, boolean, Class, Class...)}
		
		@param key The item type the node loader handles
		@param loaderClass The class name of a class that implements {@link NodeLoader},
			or null to unregister a node loader
		@throws NoSuchMethodException If the given class has no public no-argument
			constructor
		@throws ClassCastException If the given class doesn't implement the
			{@link NodeLoader} interface
	*/
	public void registerNodeLoader(Object key, Class<? extends NodeLoader<? extends BasicItem>> loaderClass)
		throws NoSuchMethodException, ClassCastException
	{
		if (loaderClass != null) ClassUtil.checkClass(loaderClass, true, NodeLoader.class);
		registerCheckedNodeLoader(key, loaderClass);
	}
	
	/**
		Register a node loader implementation that you are sure fulfills the
		requirements. Eg. it must have a public no-argument constructor and
		implement the {@link NodeLoader} interface.
		@see #registerNodeLoader(Object, Class)
	*/
	protected void registerCheckedNodeLoader(Object key, Class<? extends NodeLoader<? extends BasicItem>> loaderClass)
	{
		if (loaderClass == null)
		{
			loaders.remove(key);
		}
		else
		{
			loaders.put(key, loaderClass);
		}
	}
	
	/**
		Set a flag that determines if the factory should create 
		node loaders that automatically loads child nodes or not.
		If this flag has been set this factory will wrap the
		node loaders with a {@link AutoChildNodeLoader}.
		<p>
		
		This will generate a recursive pattern if the used node loader:
		<ul>
		<li>Uses this factory to create child node loaders. Eg. it should use
			the node loader factory provided by {@link OverviewContext#getNodeLoaderFactory()}.
			The node loader should NOT instantiate child new loaders by itself.
		<li>Uses an implementation that recurses into folder-type nodes, for example
			by extending {@link AbstractNodeLoader}.
		</ul>
		
		@param autoLoadChildren TRUE to create node loaders that automatically
			loads children, FALSE otherwise
	*/
	public void setAutoLoadChildren(boolean autoLoadChildren)
	{
		this.autoLoadChildren = autoLoadChildren;
	}
	
	/**
		Get the auto load children setting.
		@return TRUE if child nodes are automatically loaded, FALSE if not
	*/
	public boolean isAutoLoadingChildren()
	{
		return autoLoadChildren;
	}
	
	/**
		Get the null loader setting.
		@see #setUseNullLoaderIfNotFound(boolean)
	*/
	public boolean useNullLoaderIfNotFound()
	{
		return useNullLoader;
	}
	
	/**
		Set a flag that determines if a {@link NullNodeLoader} should be used
		or if an {@link ItemNotFoundException} should be thrown
		if no registered node loader is found for a specific key.
		@param useNullLoader TRUE to use a NullNodeLoader, FALSE to throw
			an exception
	 */
	public void setUseNullLoaderIfNotFound(boolean useNullLoader)
	{
		this.useNullLoader = useNullLoader;
	}
	
	/**
		Registers default item node loaders from {@link BioSource}
		to {@link Experiment}, etc.
	*/
	protected void registerDefaultNodeLoaders()
	{
		registerCheckedNodeLoader(Item.BIOSOURCE, BioSourceLoader.class);
		registerCheckedNodeLoader(Item.SAMPLE, SampleLoader.class);
		registerCheckedNodeLoader(Item.EXTRACT, ExtractLoader.class);
		registerCheckedNodeLoader(Item.ARRAYDESIGN, ArrayDesignLoader.class);
		registerCheckedNodeLoader(Item.ARRAYBATCH, ArrayBatchLoader.class);
		registerCheckedNodeLoader(Item.ARRAYSLIDE, ArraySlideLoader.class);
		registerCheckedNodeLoader(Item.PHYSICALBIOASSAY, PhysicalBioAssayLoader.class);
		registerCheckedNodeLoader(Item.DERIVEDBIOASSAY, DerivedBioAssayLoader.class);
		registerCheckedNodeLoader(Item.RAWBIOASSAY, RawBioAssayLoader.class);
		registerCheckedNodeLoader(Item.ROOTRAWBIOASSAY, RootRawBioAssayLoader.class);
		registerCheckedNodeLoader(Item.EXPERIMENT, ExperimentLoader.class);
		registerCheckedNodeLoader(Item.BIOPLATE, BioPlateLoader.class);
		registerCheckedNodeLoader(Item.BIOWELL, BioWellLoader.class);
		registerCheckedNodeLoader(Item.ITEMLIST, ItemListLoader.class);
		registerCheckedNodeLoader(Item.BIOPLATEEVENT, BioPlateEventLoader.class);
		
		// Property node loaders
		registerCheckedNodeLoader(Item.PROTOCOL, ProtocolLoader.class);
		registerCheckedNodeLoader(Item.TAG, TagLoader.class);
		registerCheckedNodeLoader("PROTOCOL.PARAMETER", ProtocolParameterLoader.class);
		registerCheckedNodeLoader("EXPERIMENTAL.FACTOR", ExperimentalFactorLoader.class);
		registerCheckedNodeLoader(Item.SOFTWARE, SoftwareLoader.class);
		registerCheckedNodeLoader(Item.HARDWARE, HardwareLoader.class);
		registerCheckedNodeLoader(Item.KIT, KitLoader.class);
		registerCheckedNodeLoader(Item.ANNOTATION, AnnotationLoader.class);
		registerCheckedNodeLoader(Item.PLATFORM, PlatformLoader.class);
		registerCheckedNodeLoader(Item.FILESETMEMBER, DataFileLoader.class);
		registerCheckedNodeLoader(Item.ANYTOANY, AnyToAnyLoader.class);
		
	}

}
