/**
	$Id: RawBioAssayLoader.java 7494 2018-06-04 06:42:46Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;


import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.listable.ExtractToParentExtractTransformer;
import net.sf.basedb.util.listable.SourceItemTransformer;
import net.sf.basedb.util.listable.TransformContext;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.NodeAttribute;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.filter.BasicItemFilter;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for raw bioassays. The forward-loading
	direction goes from {@link DerivedBioAssay}:s -&gt; {@link RawBioAssay}:s -&gt; 
	{@link Experiment}:s. The reverse-loading direction is the opposite 
	direction.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2018-06-04 08:42:46 +0200 (må, 04 jun 2018) $
*/
public class RawBioAssayLoader
	extends BasicItemNodeLoader<RawBioAssay>
{
	public RawBioAssayLoader()
	{
		super(Item.RAWBIOASSAY, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<RawBioAssay>("rawbioassay", "Raw bioassay"));
	}

	/*
		From the NodeLoader interface
		------------------------------
	*/
	@Override
	public Node createRootNode(DbControl dc, OverviewContext context, RawBioAssay item)
	{
		Node node = super.createRootNode(dc, context, item);
		try
		{
			Extract extract = item.getParentExtract();
			if (extract != null) node.setAttribute(NodeAttribute.EXTRACT, extract);
		}
		catch (PermissionDeniedException ex)
		{}
		return node;
	}
	/**
		Create forward-loading raw bioassay nodes from a given parent derived bioassay or
		extract node. The returned node is a folder-type node containing item nodes for 
		each raw bioassay. If the parent item is an extract only raw bioassays which doesn't
		have a parent derived bioassay are loaded.
		@return The folder node for the raw bioassays
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node returnNode = null;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.DERIVEDBIOASSAY)
		{
			returnNode = createForwardNode((DerivedBioAssay)parentNode.getItem(dc), dc, context, parentNode);
		}
		else if (parentType == Item.EXTRACT)
		{
			returnNode = createForwardNode((Extract)parentNode.getItem(dc), dc, context, parentNode);
		}
		else if (parentType == Item.ITEMLIST)
		{
			ItemListLoader.loadMemberNodes(this, dc, context, parentNode);
		}
		return returnNode;
	}

	/**
		Create a reverse-loading raw bioassay nodes from a given experiment node.
		The returned node is a folder-type node containing nodes for each raw
		bioassay is part of the experiment.
		@return The folder node for the raw bioassays
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node experimentNode)
	{
		NodeFactory<RawBioAssay> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		Experiment experiment = (Experiment)experimentNode.getItem(dc);
		ItemQuery<RawBioAssay> query = context.initQuery(experiment.getRawBioAssays(), "name");
		ItemResultIterator<RawBioAssay> it = query.iterate(dc);
		while (it.hasNext())
		{
			RawBioAssay rba = it.next();
			if (folderNode == null)
			{
				folderNode =  new Node("rawbioassays", "Raw bioassays", experimentNode, ChildNodeDirection.REVERSE);
			}
			Node rawNode = createItemNode(nf, rba, rba, false, folderNode, ChildNodeDirection.REVERSE);
			try
			{
				// We must set the EXTRACT attribute so that upstream loaders
				// follow the correct path when reaching the biomaterials section
				rawNode.setAttribute(NodeAttribute.EXTRACT, rba.getParentExtract());
			}
			catch (PermissionDeniedException ex)
			{}
		}
		postValidateFolder(nf, folderNode, experimentNode, false);
		return folderNode;
	}
	// ------------------------------
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Load property nodes for a raw bioassay:
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Data files: {@link DataFileLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Platform: {@link PlatformLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Software: {@link SoftwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Array design: {@link ArrayDesignLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		Extract e = getExtract((RawBioAssay)rawBioAssayNode.getItem());
		if (e == null || rawBioAssayNode.getFirstParent(new BasicItemFilter(e)) == null)
		{
			// Only load the extract if it is not already a parent node
			getNodeLoader(context, Item.EXTRACT).createReverseNode(dc, context, rawBioAssayNode);
		}
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.FILESETMEMBER).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.PLATFORM).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.ARRAYDESIGN).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.SOFTWARE).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, rawBioAssayNode);
	}
	/**
		Loads all experiment nodes where the given raw bioassay is used.
		@see ExperimentLoader#createForwardNode(DbControl, OverviewContext, Node) 
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		getNodeLoader(context, Item.EXPERIMENT).createForwardNode(dc, context, rawBioAssayNode);
	}
	/**
		Loads the derived bioassay node that this raw bioassay is created from.
		@see DerivedBioAssayLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		getNodeLoader(context, Item.DERIVEDBIOASSAY).createReverseNode(dc, context, rawBioAssayNode);
	}
	// --------------------------------

	private Node createForwardNode(DerivedBioAssay bioAssay, DbControl dc, OverviewContext context, Node bioAssayNode)
	{
		NodeFactory<RawBioAssay> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<RawBioAssay> query = context.initQuery(bioAssay.getRawBioAssays(), "name");

		// We should only load derived bioassays that has a null extract or an
		// extract that is found in the parent chain
		Set<Integer> extractIds = getExtractChain(dc, bioAssayNode);
		if (extractIds.size() > 0)
		{
			query.restrict(
				Restrictions.or(
					Restrictions.eq(Hql.property("parentExtract"), null),
					Restrictions.in(Hql.property("parentExtract"), Expressions.parameter("extracts"))
				));
			query.setParameter("extracts", extractIds, Type.INT);
		}
	
		ItemResultIterator<RawBioAssay> it = query.iterate(dc);
		while (it.hasNext())
		{
			RawBioAssay rba = it.next();
			if (folderNode == null)
			{
				folderNode =  new Node("rawbioassays", "Raw bioassays", bioAssayNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, rba, rba, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, bioAssayNode, false);
		return folderNode;
	}
	
	private Node createForwardNode(Extract extract, DbControl dc, OverviewContext context, Node extractNode)
	{
		NodeFactory<RawBioAssay> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		// Load raw bioassays linked to the 'extract' if:
		// they have no parent bioassay OR
		// the parent bioassay has no parent extract OR
		// the parent bioassay has a different parent extract
		ItemQuery<RawBioAssay> query = RawBioAssay.getQuery();
		query.restrict(Restrictions.eq(Hql.property("parentExtract"), Hql.entity(extract)));
		query.join(Hql.leftJoin("parentBioAssay", "parent"));
		query.join(Hql.leftJoin("parent", "extract", "parentExtract", null, false));
		query.restrict(
				Restrictions.or(
					Restrictions.eq(Hql.alias("parent"), null),
					Restrictions.eq(Hql.alias("parentExtract"), null),
					Restrictions.neq(Hql.alias("parentExtract"), Hql.entity(extract))
					));
		context.initQuery(query, "name");
		ItemResultIterator<RawBioAssay> it = query.iterate(dc);
		while (it.hasNext())
		{
			RawBioAssay rba = it.next();
			if (folderNode == null)
			{
				folderNode =  new Node("rawbioassays", "Raw bioassays", extractNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, rba, rba, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, extractNode, false);
		return folderNode;
	}
	
	/**
		Get the extract that is associated with the current raw bioassay
		@return The extract or null if none was found
	*/
	private Extract getExtract(RawBioAssay bioAssay)
	{
		Extract extract = null;
		try
		{
			extract = bioAssay.getParentExtract();
		}
		catch (PermissionDeniedException ex)
		{}
		return extract;
	}

	private Set<Integer> getExtractChain(DbControl dc, Node node)
	{
		List<Node> extractNodes = node.findAll(new ItemTypeFilter(Item.EXTRACT));
		Set<Integer> extractIds = new HashSet<Integer>(extractNodes.size());
		for (Node n : extractNodes)
		{
			extractIds.add(n.getItem().getId());
			if (n.getParent() == null)
			{
				// We have reach the root node, continue to load parent extracts from here
				SourceItemTransformer transformer = new ExtractToParentExtractTransformer(true);
				extractIds.addAll(transformer.transform(new TransformContext(dc), Collections.singleton(n.getItem().getId())));
				
			}
		}
		return extractIds;
	}

}
