/*
$Id: HelpExporter.java 6473 2014-06-11 13:48:39Z nicklas $

Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
Copyright (C) 2007 Martin Svensson

This file is part of BASE - BioArray Software Environment.
Available at http://base.thep.lu.se/

BASE is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

BASE is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.Client;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Help;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.Job.ExecutionTime;
import net.sf.basedb.core.plugin.AbstractExporterPlugin;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.ResultIterator;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.jdom2.DocType;
import org.jdom2.Element;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;

/**
 	This plugin exports helptexts stored in the database to a XML-file. 

 	@author Martin, Nicklas
 	@version 2.0
	@base.modified $Date$
*/
public class HelpExporter
	extends AbstractExporterPlugin
	implements InteractivePlugin, SignalTarget
{
	
	private static final GuiContext clientItem = new GuiContext(Item.CLIENT, GuiContext.Type.ITEM);
	private static final GuiContext helpList = new GuiContext(Item.HELP, GuiContext.Type.LIST);
	
	private static final Set<GuiContext> guiContexts =  
		new HashSet<GuiContext>(Arrays.asList(clientItem, helpList));

	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	
	private ItemParameterType<Client> clientType;
	private PluginParameter<Client> clientParameter;
	
	private RequestInformation configureJob;
	private ThreadSignalHandler signalHandler;
	
	private RequestInformation getConfigureJobParameters(Boolean requireFile)
	{
		if (configureJob == null)
		{
			clientType = new ItemParameterType<Client>(Client.class, null, true, 1, null);
			clientParameter = new PluginParameter<Client>(
					"client", "Client", "The client the help texts belongs to", clientType);
			
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			parameters.add(clientParameter);
			if (requireFile == null)
			{
				PluginDefinition pd = job.getPluginDefinition();
				requireFile = pd == null ? false : !pd.getAllowImmediateExecution();
			}
			String defaultPath = null;
			if (requireFile)
			{
				int currentClientId = sc.getCurrentContext(Item.CLIENT).getId();
				if (currentClientId != 0)
				{
					DbControl dc = sc.newDbControl();
					try
					{
						Client c = Client.getById(dc, currentClientId);
						defaultPath = "~/" + Path.makeSafeFilename(c.getName(), "").toLowerCase() + ".helptexts.xml";
					}
					finally
					{
						if (dc != null) dc.close();
					}
				}
			}
			parameters.add(getSaveAsParameter(null, null, defaultPath, requireFile));
			parameters.add(getOverwriteParameter(null, null));
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Help exporter options",
				"Set Client that owns the helptexts, the file path where the export file should be saved",
				parameters
			);
		}
		return configureJob;
	}
	
	/*
	 	From the InteractivePlugin interface
	 	-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}

	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		if (context.equals(clientItem))
		{
			if (item == null)
			{
				message = "The object is null";
			}
			else if (!(item instanceof Client))
			{
				message = "The object is not a Client: " + item;
			}
		}
		return message;
	}

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (Request.COMMAND_CONFIGURE_JOB.equals(command))
		{
			requestInformation = getConfigureJobParameters(null);
		}
		return requestInformation;
	}

	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(!request.isAllowedImmediateExecution());
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameters were found in the request",errors);
					return;
				}				
				
				storeValue(job, request, ri.getParameter("client"));
				if (request.getParameterValue(SAVE_AS) == null)
				{
					if (!request.isAllowedImmediateExecution())
					{
						response.setError("Immediate download is not allowed. Please specify a filename.", null);
						return;
					}
					Client client = (Client)request.getParameterValue("client");
					response.setDownloadImmediately("Export help texts for client application " + 
						client.getName(), ExecutionTime.SHORTEST, true);
				}
				else
				{
					Object parameterValue = request.getParameterValue(OVERWRITE); 
					boolean overwrite = parameterValue != null ? (Boolean)parameterValue : false;
					if (!pathCanBeUsed((String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), overwrite))
					{
						response.setError("File exists: " + (String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), null);
						return;
					}
					storeValue(job, request, ri.getParameter(SAVE_AS));
					storeValue(job, request, ri.getParameter(OVERWRITE));
					Client client = (Client)job.getValue("client");
					response.setSuggestedJobName("Export help texts for client application '" + client.getName() + "'");
					response.setDone("The job configuration is complete", ExecutionTime.SHORTEST);
				}
			}
		}
		catch(Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}		
	}
	// -------------------------------------------

	/*
	 	From the Plugin interface
	 	--------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request create and write access to File:s and Directory:s and read access to
	 	Client:s and User:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.USER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.CLIENT, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	/*
	 	From the AbstractExporterPlugin class
	 	--------------------------------------------
	*/
	private DbControl dc;
	private int numExported;
	@Override
	protected void begin(DbControl dc)
	{
		this.dc = dc;
	}
	
	@Override
	protected void performExport(ExportOutputStream out, ProgressReporter progress)
		throws IOException
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		Client client = (Client)job.getValue("client");
		client = Client.getById(dc, client.getId());
		out.setMimeType("application/xml");
		String filename = Path.makeSafeFilename(client.getName(), "").toLowerCase()+".helptexts.xml";
		out.setFilename(filename);
		out.setCharacterSet("UTF-8");

		//Query to get all helpitems for the client
		ItemQuery<Help> helpTextQuery = client.getHelp();			
		helpTextQuery.setReturnTotalCount(false);
		
		//Initialize outputstream and xmldocument
		XMLOutputter xmlOut = new XMLOutputter(Format.getPrettyFormat().setEncoding("UTF-8"));
		Writer helpExportStream = new OutputStreamWriter(out, "UTF-8");
		helpExportStream.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		xmlOut.output(new DocType("helptexts", "helpfile.dtd"), helpExportStream);
		Element helpRoot = new Element("helptexts");
		helpRoot.setAttribute("clientExternalID", client.getExternalId());
		
		ResultIterator<Help> result = null;	
		float percentDone = 0;
		numExported = 0;
		float progLastUpdated = 0;
		int updateInterval = 5;

		//Get helpitems from database
		ThreadSignalHandler.checkInterrupted();
		long numHelpItems = helpTextQuery.count(dc);
		float percentPerItem = 100.0f / numHelpItems;

		ThreadSignalHandler.checkInterrupted();
		result = helpTextQuery.iterate(dc);
		while (result.hasNext())
		{
			ThreadSignalHandler.checkInterrupted();
			//Create a new helpitem element					
			Help helpItem = result.next();
			Element newHelpElement = new Element("helpitem");
			newHelpElement.setAttribute("externalID", helpItem.getExternalId());
			
			Element name = new Element("name");
			Element description = new Element("description");
			
			name.setText(helpItem.getName());
			description.setText(helpItem.getDescription());
			
			newHelpElement.addContent(name);
			newHelpElement.addContent(description);
			
			helpRoot.addContent(newHelpElement);
			numExported++;
			percentDone = numExported*percentPerItem;
			
			if ((progress != null) && ((percentDone-progLastUpdated) > updateInterval))
			{
				progress.display((int)percentDone, numExported + " of " + numHelpItems + " helptexts have been exported");
				progLastUpdated = percentDone;
			}
		}
	
		//Finish the XML document, close the stream and commit to DB
		xmlOut.output(helpRoot, helpExportStream);
		helpExportStream.flush();
	}
	
	@Override
	protected void end(boolean success)
	{
		this.dc = null;
	}
	@Override
	protected String getSuccessMessage()
	{
		return numExported + " help texts were exported successfully";
	}
	// -------------------------------------------

	
}
