/**
	$Id: BioPlateImporter.java 7331 2017-04-10 07:51:04Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioPlateType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PlateGeometry;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.DefaultItemFunction;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing bioplate items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@since 2.16
	@base.modified $Date: 2017-04-10 09:51:04 +0200 (må, 10 apr 2017) $
*/
public class BioPlateImporter
	extends AbstractItemImporter<BioPlate>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.BIOPLATE, GuiContext.Type.LIST));

	protected static final PluginParameter<String> bioPlateTypeColumnMapping = new PluginParameter<String>(
			"bioPlateTypeColumnMapping",
			"Bioplate type",
			"Mapping that picks the name or ID of the bioplate's type from the data columns. The plug-in " +
			"will first try to find a bioplate type with the given name. If none is found and the value " +
			"is numeric it will try to load by internal ID. Required when creating new items. " +
			"Example: \\Bioplate type\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> plateGeometryColumnMapping = new PluginParameter<String>(
			"plateGeometryColumnMapping",
			"Plate geometry",
			"Mapping that picks the name or ID of the bioplate's geometry from the data columns. The plug-in " +
			"will first try to find a plate geometry with the given name. If none is found and the value is " +
			"numeric it will try to load by internal ID. Required when creating new items. " +
			"Example: \\Plate geometry\\",
			optionalColumnMapping
			);

	
	protected static final PluginParameter<String> hardwareColumnMapping = new PluginParameter<String>(
			"hardwareColumnMapping",
			"Storage location",
			"Mapping that picks the name or ID of the storage location where the bioplate is stored " +
			"from the data columns. The plug-in will first try to find a location with the given name. " +
			"If none is found and the value is numeric it will try to load by internal ID. " +
			"Example: \\Freezer\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> sectionColumnMapping = new PluginParameter<String>(
			"sectionColumnMapping",
			"Section",
			"Mapping that picks the section where the bioplate is stored from the data columns. This " +
			"parameter is optional. Example: \\Section\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> trayColumnMapping = new PluginParameter<String>(
			"trayColumnMapping",
			"Tray",
			"Mapping that picks the tray within a section where the bioplate is stored from the data columns. This " +
			"parameter is optional. Example: \\Tray\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> positionColumnMapping = new PluginParameter<String>(
			"positionColumnMapping",
			"Position",
			"Mapping that picks the position within a tray where the bioplate is stored from the data columns. This " +
			"parameter is optional. Example: \\Position\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> barcodeColumnMapping = new PluginParameter<String>(
			"barcodeColumnMapping",
			"Barcode",
			"Mapping that picks the barcode of the plate from the data columns. This " +
			"parameter is required if the 'Barcode' identification method is selected, " +
			"otherwise it is optional. Example: \\Barcode\\",
			optionalColumnMapping
			);


	public BioPlateImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper bioPlateTypeMapper;
	private Mapper plateGeometryMapper;
	private Mapper createdMapper;
	private Mapper registeredMapper;
	private Mapper hardwareMapper;
	private DefaultItemFunction<Hardware> defaultHardwareFunction;
	private Mapper sectionMapper;
	private Mapper trayMapper;
	private Mapper positionMapper;
	private Mapper barcodeMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link PropertyIdMethod#BARCODE}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				PropertyIdMethod.BARCODE,
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	/**
		Adds column mappings for name, description, plate type, plate
		geometry, barcode and freezer.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(bioPlateTypeColumnMapping);
		parameters.add(plateGeometryColumnMapping);
		parameters.add(SampleImporter.createdColumnMapping);
		parameters.add(registeredColumnMapping);
		parameters.add(barcodeColumnMapping);
		parameters.add(hardwareColumnMapping);
		parameters.add(sectionColumnMapping);
		parameters.add(trayColumnMapping);
		parameters.add(positionColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Calls {@link BioPlate#getQuery()}.
	*/
	@Override
	protected ItemQuery<BioPlate> createItemQuery()
	{
		return BioPlate.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? BioPlate.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? BioPlate.MAX_DESCRIPTION_LENGTH : null, null);
		bioPlateTypeMapper = getMapper(ffp, (String)job.getValue("bioPlateTypeColumnMapping"), null, null);
		plateGeometryMapper = getMapper(ffp, (String)job.getValue("plateGeometryColumnMapping"), null, null);
		createdMapper = getMapper(ffp, (String)job.getValue("createdColumnMapping"), null, null);
		registeredMapper = getMapper(ffp, (String)job.getValue("registeredColumnMapping"), null, null);
		defaultHardwareFunction = getDefaultItemFunction(Item.HARDWARE);
		hardwareMapper = getMapper(ffp, (String)job.getValue("hardwareColumnMapping"), null, null, defaultHardwareFunction);
		sectionMapper = getMapper(ffp, (String)job.getValue("sectionColumnMapping"), 
				cropStrings ? BioPlate.MAX_SECTION_LENGTH : null, null);
		trayMapper = getMapper(ffp, (String)job.getValue("trayColumnMapping"), 
				cropStrings ? BioPlate.MAX_TRAY_LENGTH : null, null);
		positionMapper = getMapper(ffp, (String)job.getValue("positionColumnMapping"), 
				cropStrings ? BioPlate.MAX_POSITION_LENGTH : null, null);
		barcodeMapper = getMapper(ffp, (String)job.getValue("barcodeColumnMapping"), 
				cropStrings ? BioPlate.MAX_BARCODE_LENGTH : null, null);
	}
	
	@Override
	protected BioPlate createItem(DbControl dc, FlatFileParser.Data data)
	{
		BioPlateType plateType = null;
		PlateGeometry geometry = null;
		if (bioPlateTypeMapper != null) 
		{
			plateType = findBioPlateType(dc, FallbackIdMethod.NAME_OR_ID, bioPlateTypeMapper.getValue(data));
		}
		if (plateGeometryMapper != null)
		{
			geometry = findPlateGeometry(dc, FallbackIdMethod.NAME_OR_ID, plateGeometryMapper.getValue(data));
		}

		BioPlate plate = BioPlate.getNew(dc, geometry, plateType);
		if (registeredMapper != null) 
		{
			plate.setEntryDate(parseDate(registeredMapper.getValue(data)));
		}
		updateItem(dc, plate, data);
		return plate;
	}
	
	@Override
	protected void updateItem(DbControl dc, BioPlate plate, FlatFileParser.Data data)
	{
		if (nameMapper != null) plate.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) plate.setDescription(descriptionMapper.getValue(data));
		if (createdMapper != null) plate.setEventDate(parseDate(createdMapper.getValue(data)));
		if (barcodeMapper != null) plate.setBarcode(barcodeMapper.getValue(data));
		if (hardwareMapper != null)
		{
			ItemSubtype type = ItemSubtype.getById(dc, SystemItems.getId(Hardware.FREEZER));
			defaultHardwareFunction.setCurrentSubtype(type);
			String nameOrId = hardwareMapper.getValue(data);
			Hardware freezer = findHardware(dc, FallbackIdMethod.NAME_OR_ID, nameOrId, type);
			if (nameOrId == null || freezer != null) plate.setFreezer(freezer);
		}
		if (sectionMapper != null) plate.setSection(sectionMapper.getValue(data));
		if (trayMapper != null) plate.setTray(trayMapper.getValue(data));
		if (positionMapper != null) plate.setPosition(positionMapper.getValue(data));
	}
	// ------------------------------------
}
