/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.executor;

import java.io.FileFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.File;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.User;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.ChainedProgressReporter;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.importer.DiskFileWrapper;
import net.sf.basedb.util.importer.spotdata.BaseFileImporter;

/**
	Importer implementation for the external program executor that
	imports data from BASEfile format.
	
	@author nicklas
	@since 2.16
	@base.modified $Date $
*/
public class BaseFileImporterSupport
	extends AbstractIOSupport
{
	
	public BaseFileImporterSupport()
	{}

	/*
		From the Plugin interface
		-------------------------
	*/
	@Override
	public MainType getMainType()
	{
		return Plugin.MainType.IMPORT;
	}
	// -------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		try
		{
			String command = request.getCommand();
			if (command.equals(IOSupport.COMMAND_CONFIGURE_IO_PLUGIN))
			{
				RequestInformation ri = getConfigureExecutorParameters(command);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found " +
														"in the request", errors);
					return;
				}
				
				// Import options
				storeValue(configuration, request, ri.getParameter("datafile"));
				storeValue(configuration, request, ri.getParameter("extraFiles"));

				response.setDone("BASEfile importer configuration complete");
			}
			else if (command.equals(IOSupport.COMMAND_CONFIGURE_IO_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(command);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found " +
														"in the request", errors);
					return;
				}
			
				storeValue(job, request, ri.getParameter(Parameters.CHILD_NAME));
				storeValue(job, request, ri.getParameter(Parameters.CHILD_DESCRIPTION));
				storeValue(job, request, ri.getParameter(Parameters.TRANSFORMATION_NAME));
				storeValue(job, request, ri.getParameter("extraFilesDirectory"));
				response.setDone("BASEfile importer configuration complete");
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(IOSupport.COMMAND_CONFIGURE_IO_PLUGIN))
		{
			requestInformation = getConfigureExecutorParameters(command);
		}
		else if (command.equals(IOSupport.COMMAND_CONFIGURE_IO_JOB))
		{
			requestInformation = getConfigureJobParameters(command);
		}
		return requestInformation;
	}
	// -----------------------------------------
	/*
		From the IOSupport interface
		----------------------------
	*/
	@Override
	public String getShortName()
	{
		return "BASEfile importer";
	}
	// ----------------------------

	@Override
	protected String performIO(DbControl dc, ProgressReporter progress)
		throws IOException
	{
		// Get job configuration parameters
		ExternalProgramExecutor executor = getExecutor();
		BioAssaySet source = executor.getCurrentBioAssaySet(dc);
		String datafile = Values.getString((String)configuration.getValue("datafile"), "stdout.txt");
		String transformationName = (String)job.getValue(Parameters.TRANSFORMATION_NAME);
		String childName = (String)job.getValue(Parameters.CHILD_NAME);
		String childDescription = (String)job.getValue(Parameters.CHILD_DESCRIPTION);
		String extraFileDirectory = (String)job.getValue("extraFilesDirectory");
		ChainedProgressReporter chainedProgress = progress == null ? 
				null : new ChainedProgressReporter(progress);
		
		// Create child transformation
		Job currentJob = getCurrentJob(dc);
		Transformation t = source.newTransformation(currentJob);
		t.setName(transformationName);
		dc.saveItem(t);

		// Create the importer
		if (chainedProgress != null) chainedProgress.setRange(0, 90);
		BaseFileImporter importer = new BaseFileImporter();
		importer.useBase2ColumnNames(source.getRawDataType().getChannels());
		importer.setDbControl(dc);
		importer.setProgressReporter(chainedProgress);
		java.io.File in = executor.createFile(datafile);
		importer.setSourceFileWrapper(new DiskFileWrapper(in));
		importer.setTransformation(t);
		BioAssaySet child = importer.doImport();
		if (child != null)
		{
			child.setName(childName);
			child.setDescription(childDescription);
		}
		
		// Import extra files
		if (extraFileDirectory != null)
		{
			final List<java.io.File> extraFiles = executor.inspectWorkingDirectory();
			extraFiles.remove(in);
			if (extraFiles.size() > 0)
			{
				Directory extraDir = getExtraFilesDirectory(dc, extraFileDirectory);
				FileFilter newFiles = new FileFilter()
				{
					@Override
					public boolean accept(java.io.File pathname)
					{
						return pathname.isDirectory() || extraFiles.contains(pathname);
					}
				};
				if (chainedProgress != null) chainedProgress.setRange(90, 100);
				List<File> uploaded = FileUtil.uploadFiles(dc, extraDir, executor.getWorkingDirectory(), 
						newFiles, false, chainedProgress);
				BasicItem from = child == null ? t : child;
				AnyToAny ata = AnyToAny.getNew(dc, from, extraDir, "./", false);
				dc.saveItem(ata);
				for (File f : uploaded)
				{
					Directory d = f.getDirectory();
					if (d == extraDir)
					{
						// Attach file to child bioassay set or transformation
						ata = AnyToAny.getNew(dc, from, f, "./" + f.getName(), false);
						dc.saveItem(ata);
					}
					else if (d.getParent() == extraDir)
					{
						// Attach directory to child bioassay set or transformation
						ata = AnyToAny.getNew(dc, from, d, "./" + d.getName(), false);
						dc.saveItem(ata);
					}
				}
			}
		}

		String msg = "Import complete";
		if (child != null)
		{
			msg = "Bioassay set '" + child.getName() + "' successfully created";
		}
		else 
		{
			msg = "Transformation '" + t.getName() + "' successfully created";
		}
		return msg;
	}

	private RequestInformation configureExecutor;

	private RequestInformation getConfigureExecutorParameters(String command)
	{
		if (configureExecutor != null) return configureExecutor;
	
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			// Name of the metadata file
			parameters.add(new PluginParameter<String>(
				"datafile", 
				"Data filename", 
				"The name of the output file from the external program. The defult value is 'stdout.txt' which is " +
				"automatically generated if the program writes to standard output. It is possible " +
				"to use other filenames but the program is then responsible to create the file itself.",
				new StringParameterType(255, "stdout.txt", true))
			);
			
			parameters.add(new PluginParameter<Boolean>(
				"extraFiles", 
				"Enable extra file import", 
				"Select this option if the external program generates extra files and you " +
				"want the user to be able to import those files to BASE. ", 
				new BooleanParameterType()));
						
			configureExecutor = new RequestInformation(
					command,
					"Select import options.",
					"Specify options for the import phase of the program execution.",
					parameters
				);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureExecutor;
	}

	private RequestInformation configureJob;

	private RequestInformation getConfigureJobParameters(String command)
	{
		if (configureJob != null) return configureJob;
	
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			String transformationName = configuration.getPluginConfiguration().getName();
			parameters.add(Parameters.transformationNameParameter(null, null, transformationName));
			parameters.add(Parameters.childNameParameter(null, null, "After " + transformationName));
			parameters.add(Parameters.childDescriptionParameter(null, null, null));
			
			if (Boolean.TRUE.equals(configuration.getValue("extraFiles")))
			{
				Experiment experiment = getExecutor().getCurrentBioAssaySet(dc).getExperiment();
				String homeDirectoryPath = "/" + Path.makeSafeFilename(transformationName, "");
				Directory homeDirectory = experiment == null ? null : experiment.getDirectory();
				if (homeDirectory != null)
				{
					homeDirectoryPath = homeDirectory.getPath() + homeDirectoryPath;
				}
				else
				{
					homeDirectory = User.getById(dc, sc.getLoggedInUserId()).getHomeDirectory();
					if (homeDirectory != null)
					{
						homeDirectoryPath = homeDirectory.getPath() + homeDirectoryPath;
					}
				}
				parameters.add(new PluginParameter<String>(
					"extraFilesDirectory",
					"Import extra files to",
					"Select a directory where all extra files generated by " +
					"the program are saved. If no directory is selected, the extra " +
					"files are not imported to BASE.",
					new PathParameterType(Path.Type.DIRECTORY, homeDirectoryPath, false)
				));
			}
			
			configureJob = new RequestInformation(
					command,
					"Select import options.",
					"",
					parameters
				);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
	
	private Directory getExtraFilesDirectory(DbControl dc, String pathName)
	{ 
		Path path = new Path(pathName, Path.Type.DIRECTORY);
		Directory dir = null;
		try
		{
			// Load existing...
			dir = Directory.getByPath(dc, path);
		}
		catch(ItemNotFoundException e)
		{
			// Create a new directory if no existing were found
			dir = Directory.getNew(dc, path);
			dc.saveItem(dir);
		}
		return dir;
	}

	
}
