/*
	$Id: TestPluginDefinition.java 6433 2014-03-14 08:04:02Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.plugin.Plugin.MainType;
import net.sf.basedb.core.query.*;
import net.sf.basedb.util.ConsoleProgressReporter;

import java.util.EnumSet;
import java.util.Map;
import java.util.Set;

public class TestPluginDefinition
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing plugin definition");
		write_header();
		// Standard tests: create, load, list
		int pluginType1 = TestPluginType.test_create("NullPluginInterface", null, false);
		int pluginType2 = TestPluginType.test_create("JarPluginInterface", "JarPlugin.jar", false);
		
		ok = ok & TestPluginType.ok;
		
		int id = test_create("NullPlugin", null);
		int id2 = test_create("JarPlugin", "JarPlugin.jar");
		int id3 = test_create("PermissionPlugin", null);
		test_load(id);
		test_count(Item.REPORTER, GuiContext.Type.LIST);
		test_list();

		int jobId = test_execute(id2, "foo", false);
		
		// extra test: test permissions
		test_set_permissions(id3, Item.NEWS, true, null, null);
		int job1Id = test_execute(id3, "create", true);
		test_set_permissions(id3, Item.NEWS, true, EnumSet.of(Permission.CREATE), EnumSet.of(Permission.READ));
		int job2Id = test_execute(id3, "create", false);
		int job3Id = test_execute(id3, "export", true);
		test_set_permissions(id3, Item.NEWS, true, EnumSet.of(Permission.READ), EnumSet.of(Permission.WRITE));
		int job4Id = test_execute(id3, "export", false);
		test_list_permissions(id3);
		
		// Standard test: Delete
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		
		TestJob.test_delete(jobId);
		TestJob.test_delete(job1Id);
		TestJob.test_delete(job2Id);
		TestJob.test_delete(job3Id);
		TestJob.test_delete(job4Id);
		
		test_delete(id);
		test_delete(id2);
		test_delete(id3);
		
		TestPluginType.test_delete(pluginType1);
		TestPluginType.test_delete(pluginType2);
		write("++Testing plugin definition "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(String className, String jarPath)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.PLUGINDEFINITION)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition pd = PluginDefinition.getNew(dc, className, jarPath, false);
			dc.saveItem(pd);
			dc.commit();
			id = pd.getId();
			write_item(0, pd);
			write("--Create plugin definition OK");
		}
		catch (Throwable ex)
		{
			write("--Create plugin definition FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition pd = PluginDefinition.getById(dc, id);
			write_item(0, pd);
			write("--Load plugin definition OK");
		}
		catch (Throwable ex)
		{
			write("--Load plugin definition FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_count(Item itemType, GuiContext.Type contextType)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Map<MainType, Integer> count = PluginDefinition.countPlugins(dc, new GuiContext(itemType, contextType));
			for (Map.Entry<MainType, Integer> entry : count.entrySet())
			{
				if (!TestUtil.getSilent()) write(entry.getKey() + "\t" + entry.getValue());
			}
			write("--Count plugin definitions OK");
		}
		
		catch (Throwable ex)
		{
			write("--Count plugin definitions FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}
	
	static void test_list()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemQuery<PluginDefinition> q = PluginDefinition.getQuery();
			ItemResultList<PluginDefinition> l = q.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			write("--List plugin definitions OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List plugin definitions FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static int test_get(String className)
	{
		DbControl dc = null;
		int pluginDefinitionId = 0;
		try
		{
			dc = TestUtil.getDbControl();
			ItemQuery<PluginDefinition> q = PluginDefinition.getQuery();
			q.restrict(
				Restrictions.eq(
					Hql.property("className"), 
					Expressions.parameter("className")
				)
			);
			q.setParameter("className", className, Type.STRING);
			ItemResultList<PluginDefinition> pluginDefinitions = q.list(dc);
			if (pluginDefinitions.size() > 0)
			{
				PluginDefinition pd = pluginDefinitions.get(0);
				pluginDefinitionId = pd.getId();
				write("--Get plugin definition "+className+" OK");
			}
			else
			{
				write("--Get plugin definition "+className+" NOT FOUND");
			}
		}
		catch (Throwable ex)
		{
			write("--Get plugin definition FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return pluginDefinitionId;
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition pd = PluginDefinition.getById(dc, id);
			dc.deleteItem(pd);
			Set<ItemProxy> using = pd.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(pd + " is used by " + using);
			}
			dc.commit();
			write("--Delete plugin definition OK");
		}
		catch (Throwable ex)
		{
			write("--Delete plugin definition FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static int test_execute(int pluginId, String cmd, boolean shouldFail)
	{
		if (pluginId == 0) return 0;
		DbControl dc = null;
		int jobId = 0;
		try
		{
			dc = TestUtil.getDbControl();

			PluginDefinition plugin = PluginDefinition.getById(dc, pluginId);
			int numNews = (int)News.getQuery().count(dc);
			Job job = Job.getNew(dc, plugin, null, null);
			job.setParameterValue("cmd", new StringParameterType(), cmd);
			job.setParameterValue("shouldFail", new BooleanParameterType(), shouldFail);
			job.setParameterValue("numNews", new IntegerParameterType(), numNews);
			job.setScheduled(null, null);
			dc.saveItem(job);
			dc.commit();
			jobId = job.getId();
			
			dc = TestUtil.getDbControl();
			job = Job.getById(dc, jobId);
			PluginExecutionRequest request = 
				job.execute(TestUtil.getSilent() ? null : new ConsoleProgressReporter(), TestUtil.getLocalIp());
			dc.commit();
			
			PluginResponse response = request.invoke();
			if (response.getStatus() != Response.Status.DONE)
			{
				throw new BaseException(response.getMessage(), response.getErrorList() != null ?
						response.getErrorList().get(0) : null);
			}
			write("--Execute plugin with permissions OK; response=" + response.getMessage());
		}
		catch (Throwable ex)
		{
			write("--Execute plugin with permissions FAILED");
			ok = false;
			ex.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return jobId;
	}
	
	static void test_set_permissions(int id, Item itemType, boolean usePermission, Set<Permission> granted, Set<Permission> denied)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition pd = PluginDefinition.getById(dc, id);
			RoleKey rk = RoleKey.getById(dc, SystemItems.getRoleKeyId(itemType));
			pd.setPermissions(rk, granted, denied);
			pd.setUsePermissions(usePermission);
			dc.commit();			
			write("--Set plugin permission OK");
		}
		catch (Throwable ex)
		{			
			write("--Set plugin permission FAILED");
			ok = false;
			ex.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_list_permissions(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition pd = PluginDefinition.getById(dc, id);
			ItemQuery<RoleKey> q = RoleKey.getQuery();
			ItemResultList<RoleKey> l = q.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				RoleKey key = l.get(i);
				write_item(i, key, pd.getGranted(key), pd.getDenied(key));
			}
			write("--List plugin permissions OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{			
			write("--List plugin permissions FAILED");
			ok = false;
			ex.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tVersion\tCopyright\tContact\tEmail\tUrl");
			write("-- \t-- \t--------- \t-----------\t-------\t---------\t-------\t-----\t---");
		}
	}
	static void write_item(int i, PluginDefinition pd)
		throws BaseException
	{
		if (!TestUtil.getSilent())
		{
			System.out.println(i+":\t"+pd.getId()+"\t"+pd.getName()+"\t"+pd.getDescription()+
			"\t"+pd.getVersionString()+"\t"+pd.getCopyright()+"\t"+
			pd.getContact()+"\t"+pd.getEmail()+"\t"+pd.getUrl());
		}
	}
	
	static void write_item(int i, RoleKey key, Set<Permission> granted, Set<Permission> denied)
	{
		if (!TestUtil.getSilent())
		{
			System.out.println(i+":\t"+key.getName()+"\t"+granted+"\t"+denied);
		}
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}

}

