/*
	$Id: TestReporter.java 7516 2018-11-02 08:45:34Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.parser.FlatFileParser;

import java.util.regex.Pattern;
import java.util.Date;

public class TestReporter
{
	static boolean ok = true;

	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing reporter batching");
		int rtId = TestReporterType.test_create();
		test_create(50, rtId, 0, true);
		test_create(50, rtId, 50, false);
		test_update("Batchtest0");
		write_header();
		test_list(100);
		test_externalId("Batchtest0");
		
		test_import_from_file("data/test.rawdata.import.txt", "\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", "\\t", 4, 3);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete();
		TestReporterType.test_delete(rtId);
		write("++Testing reporter batching "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}
	
	static void test_create(int nbr, int rtId, int startNbr, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.REPORTER)) return;
		write("--Start reporter batching(" + nbr + ")");
		DbControl dc = null;
		long time;
		try
		{
			dc = TestUtil.getDbControl();
			ReporterType rt = ReporterType.getById(dc, rtId);
			ReporterBatcher rb = ReporterBatcher.getNew(dc);
			
			rb.setBatchSize(500);
			int toNbr = nbr + startNbr;
			time = System.currentTimeMillis();
			for (int i = startNbr; i < toNbr; ++i)
			{
				ReporterData rd = Reporter.getNew("Batchtest" + i);
				Reporter.setReporterType(rd, rt);
				if (setAll)
				{
					rd.setName("Test reporter");
					rd.setDescription("Added at "+new Date());
					rd.setSymbol("Some_symbol");
					rd.setExtended("species", "Homo Sapiens");
					rd.setExtended("sequence", "atcatcgagaggaatgagacgtgacgtagg");
					rd.setExtended("length", Integer.valueOf(30));
				}
				rb.insert(rd);
			}
			dc.commit();
			time = System.currentTimeMillis()-time;
			write("--Batching reporters OK (time="+time+" ms)");
		}
		catch (Throwable ex)
		{
			write("--Batching reporters FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static boolean test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			DataQuery<ReporterData> query = Reporter.getQuery();
			DataResultIterator<ReporterData> it = query.iterate(dc);
			int counter = 0;
			while (it.hasNext())
			{
				write_item(dc, it.next());
				++counter;
			}
			if (expectedResults >= 0 && expectedResults != counter)
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+counter);
			}
			write("--List reporters OK (" + counter + ")");
		}
		catch (Throwable ex)
		{
			write("--List reporters FAILED");
			ex.printStackTrace();
			ok = false;
			return false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return true;
	}
	
	static void test_update(String externalId)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ReporterBatcher rb = ReporterBatcher.getNew(dc);
			ReporterData rd = Reporter.getByExternalId(dc, externalId);
			rd.setExtended("species", "Mus Musculus");
			rb.update(rd);
			dc.commit();
			write("--Update by external id OK");
		}
		catch (Throwable ex)
		{
			write("--Update by external id FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static int test_externalId(String externalId)
	{
		DbControl dc = null;
		int id = 0;
		try
		{
			dc = TestUtil.getDbControl();
			ReporterData rd = Reporter.getByExternalId(dc, externalId);
			id = rd.getId();
			write_item(dc, rd);
			write("--Get by external id OK");
		}
		catch (Throwable ex)
		{
			write("--Get by external id FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_delete()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ReporterBatcher rb = ReporterBatcher.getNew(dc);
			DataQuery<ReporterData> query = Reporter.getQuery();
			DataResultIterator<ReporterData> it = query.iterate(dc);
			while (it.hasNext())
			{
				rb.delete(it.next());
			}
			dc.commit();
			write("--Delete reporters OK");
		}
		catch (Throwable ex)
		{
			write("--Delete reporters FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("ID \tExternal ID\tName      \tDescription  \tSymbol    \tLast update   \tCluster ID  \tReporter type");
			write("-- \t-----------\t--------- \t-------------\t----------\t--------------\t---------\t-------------");
		}
	}
	static void write_item(DbControl dc, ReporterData rd)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(rd.getId()+"\t"+rd.getExternalId()+"\t"+rd.getName()+"\t"+rd.getDescription()+
			"\t"+rd.getSymbol()+"\t"+rd.getLastUpdate()+"\t"+rd.getExtended("clusterId")+"\t"+Reporter.getReporterType(dc, rd));
	}
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void test_import_from_file(String filename, String headerRegexp, String splitterRegexp, int idCol, int nameCol)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.REPORTER)) return;
		write("--Start import reporters from file (" + filename + ")");
		DbControl dc = null;
		long time;
		int i = 0;
		try
		{
			dc = TestUtil.getDbControl();

			FlatFileParser ffp = new FlatFileParser();
			ffp.setDataHeaderRegexp(Pattern.compile(headerRegexp));
			ffp.setDataSplitterRegexp(Pattern.compile(splitterRegexp));
			ffp.setInputStream(FileUtil.getInputStream(new java.io.File(filename)), "ISO-8859-1");
			ffp.parseHeaders();
			ReporterBatcher rb = ReporterBatcher.getNew(dc);
			time = System.currentTimeMillis();
			while (ffp.hasMoreData())
			{
				FlatFileParser.Data data = ffp.nextData();
				String externalId = data.get(idCol);
				String name = data.get(nameCol);
				if (name == null) name = externalId;
				if (externalId != null && !rb.exists(externalId, true, true))
				{
					ReporterData r = Reporter.getNew(externalId);
					r.setName(name);
					rb.insert(r);
					i++;
				}
			}
			dc.commit();
			time = System.currentTimeMillis()-time;
			write("--Import reporters from file OK ("+i+", time="+time+" ms)");
		}
		catch (Throwable ex)
		{
			write("--Import reporters from file FAILED ("+i+")");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}

	}
	
}

