/**
	$Id: StringMatcher.java 4515 2008-09-11 20:08:14Z jari $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.fuzzy;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import com.wcohen.ss.Level2JaroWinkler;
import com.wcohen.ss.api.StringDistance;

/**
	A wrapper class for fuzzy string matching using the SecondString package. This
	class uses a given StringDistance object for calculating string similaroties.
	Use {@link #getScore(String, String)} to get the similarity score of two strings.
	<p>
	Use {@link #getBestMatch(String, Collection)} to find the best match of a string among
	strings in a given list.
	<p>
	Use {@link #getBestPairs(Collection, Collection)} to pair up strings in two lists.

	@see <a href="http://secondstring.sourceforge.net/">SecondString project page</a>
	@author nicklas
	@version 2.8
	@base.modified $Date: 2008-09-11 22:08:14 +0200 (to, 11 sep 2008) $
*/
public class StringMatcher
{

	private final StringDistance matcher;
	
	/**
		Create a new matcher using the default fuzzy matching algorithm:
		{@link Level2JaroWinkler}.
	*/
	public StringMatcher()
	{
		this(new Level2JaroWinkler());
	}
	
	/**
		Create a new matcher using a specific fuzzy matching algorithm.
		@param matcher The algorithm to use
	*/
	public StringMatcher(StringDistance matcher)
	{
		this.matcher = matcher;
	}
	
	/**
		Get the similarity score of two strings. The score is a value between 0
		and 1, where 0 is a poor match and 1 is a good match. Note! It doesn't
		have to be a perfect match to get a score of 1.
		@param s1 The first string
		@param s2 The second string
		@return The similarity score or 0 if any of the strings are null
	*/
	public double getScore(String s1, String s2)
	{
		if (s1 == null || s2 == null) return 0.0;
		return matcher.score(s1, s2);
	}
	
	/**
		Find the string that is most similar to a given string in a list of
		strings.
		@param key The string to look for
		@param values The list of strings to compare with the key
		@return A {@link FuzzyMatch} result for the string in the list
			that got the highest score; null if there are no values in the
			list
	*/
	public FuzzyMatch getBestMatch(String key, Collection<? extends String> values)
	{
		if (values == null || values.size() == 0) return null;
		double bestScore = Double.MIN_NORMAL;
		String bestMatch = null;
		int i = 0;
		for (String s : values)
		{
			double score = getScore(key, s);
			if (score > bestScore)
			{
				bestScore = score;
				bestMatch = s;
			}
			++i;
		}
		return new FuzzyMatch(key, bestMatch, bestScore);
	}
	
	/**
		Match strings in two lists. The result is a paired list of strings
		with one values from each of the lists. The matching is done so that
		no string from any of the lists is paired more than once.
		
		@param keys The list with the keys
		@param values The list with the values
		@return A list of {@link FuzzyMatch} results. The returned list is of the
			same size as the 'keys' collection with the elements in the same
			order as returned by the iterator. The list may contains null elements
			if no match could be found for a given key
	*/
	public List<FuzzyMatch> getBestPairs(Collection<? extends String> keys, Collection<? extends String> values)
	{
		// First, calculate all pair-wise scores
		double[][] scores = new double[keys.size()][values.size()];
		String[] theKeys = new String[keys.size()];
		String[] theValues = new String[values.size()];
		int keyIdx = 0;
		int valueIdx = 0;
		for (String key : keys)
		{
			theKeys[keyIdx] = key;
			for (String value : values)
			{
				if (keyIdx == 0) theValues[valueIdx] = value;
				scores[keyIdx][valueIdx] = getScore(key, value);
				valueIdx++;
			}
			keyIdx++;
			valueIdx = 0;
		}
		
		FuzzyMatch[] matched = new FuzzyMatch[keys.size()];
		// Find the position with the highest score
		int[] bestIdx = getHighestScoreIdx(scores);
		while (bestIdx != null)
		{
			keyIdx = bestIdx[0];
			valueIdx = bestIdx[1];
			matched[keyIdx] = new FuzzyMatch(theKeys[keyIdx], theValues[valueIdx], scores[keyIdx][valueIdx]);
			// Clear the scores for the used key/value so it is not matched again
			for (int i = 0; i < scores[keyIdx].length; ++i)
			{
				scores[keyIdx][i] = -Double.MAX_VALUE;
			}
			for (int i = 0; i < scores.length; ++i)
			{
				scores[i][valueIdx] = -Double.MAX_VALUE;
			}
			// Get the next remaining position with the highest score
			bestIdx = getHighestScoreIdx(scores);
		}
		return Arrays.asList(matched);
	}

	private int[] getHighestScoreIdx(double[][] score)
	{
		double maxScore = -Double.MAX_VALUE;
		int maxKeyIdx = -1;
		int maxValueIdx = -1;
		int keyIdx = 0;
		int valueIdx = 0;
		while (keyIdx < score.length)
		{
			while (valueIdx < score[keyIdx].length)
			{
				if (score[keyIdx][valueIdx] > maxScore)
				{
					maxScore = score[keyIdx][valueIdx];
					maxKeyIdx = keyIdx;
					maxValueIdx = valueIdx;
				}
				valueIdx++;
			}
			keyIdx++;
			valueIdx = 0;
		}
		return maxKeyIdx >= 0 ? new int[] { maxKeyIdx, maxValueIdx } : null;
	}
	
	/**
		Wrapper that holds information about a fuzzy match.
	*/
	public static class FuzzyMatch
	{
		private final String key;
		private final String value;
		private final double score;
		
		FuzzyMatch(String key, String value, double score)
		{
			this.key = key;
			this.value = value;
			this.score = score;
		}
		
		/**
			Get the key string.
		*/
		public String getKey()
		{
			return key;
		}
		/**
			Get the value string.
		*/
		public String getValue()
		{
			return value;
		}
		/**
			Get the similarity score of the key/value strings.
			@see #getScore(String, String)
		*/
		public double getScore()
		{
			return score;
		}
	}
	
}
